# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Akini Ross (@akinross) <akinross@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

from ansible_collections.cisco.mso.plugins.module_utils.constants import TEMPLATE_TYPES, DOMAIN_TYPE_MAP, VM_DOMAIN_PROVIDER_MAP
from ansible_collections.cisco.mso.plugins.module_utils.utils import generate_api_endpoint
from collections import namedtuple

KVPair = namedtuple("KVPair", "key value")
Item = namedtuple("Item", "index details")
SearchQuery = namedtuple("SearchQuery", "key kv_pairs")


class MSOTemplate:
    def __init__(self, mso_module, template_type=None, template_name=None, template_id=None, schema_name=None, schema_id=None, fail_module=False):
        self.mso = mso_module
        self.templates_path = "templates"
        self.summaries_path = "{0}/summaries".format(self.templates_path)
        self.template = {}
        self.template_path = ""
        self.template_name = template_name
        self.template_id = template_id
        self.template_type = template_type
        self.template_summary = {}
        self.template_objects_cache = {}
        self.schema_path = None
        self.schema_name = schema_name
        self.schema_id = schema_id

        if template_id:
            # Checking if the template with id exists to avoid error: MSO Error 400: Template ID 665da24b95400f375928f195 invalid
            self.template_summary = self.mso.get_obj(self.summaries_path, templateId=self.template_id)
            if self.template_summary:
                self.template_path = "{0}/{1}".format(self.templates_path, self.template_id)
                self.template = self.mso.query_obj(self.template_path)
                self.template_name = self.template.get("displayName")
                self.template_type = self.template.get("templateType")
                self._set_deployment_properties()
                if self.template_type == "application":
                    self._set_schema_properties()
            else:
                self.mso.fail_json(
                    msg="Provided template id '{0}' does not exist. Existing templates: {1}".format(
                        self.template_id,
                        [
                            "Template '{0}' with id '{1}'".format(template.get("templateName"), template.get("templateId"))
                            for template in self.mso.query_objs(self.summaries_path)
                        ],
                    )
                )
        elif template_name:
            if not template_type:
                self.mso.fail_json(msg="Template type must be provided when using template name.")
            self.template_summary = self.mso.get_obj(
                self.summaries_path,
                templateName=self.template_name,
                templateType=TEMPLATE_TYPES[template_type]["template_type"],
                schemaName=self.schema_name,
                schemaId=self.schema_id,
            )
            if self.template_summary:
                self.template_path = "{0}/{1}".format(self.templates_path, self.template_summary.get("templateId"))
                self.template = self.mso.query_obj(self.template_path)
                self.template_id = self.template.get("templateId")
                self.template_type = self.template.get("templateType")
                self._set_deployment_properties()
                if template_type == "application":
                    self._set_schema_properties()

            if fail_module and not self.template:
                self.mso.fail_json(
                    msg="Provided template name '{0}' does not exist. Existing templates: {1}".format(
                        self.template_name,
                        [
                            "Template '{0}' with id '{1}'".format(template.get("templateName"), template.get("templateId"))
                            for template in self.mso.query_objs(self.summaries_path, templateType=TEMPLATE_TYPES[template_type]["template_type"])
                        ],
                    )
                )

        elif template_type:
            self.template = self.mso.query_objs(self.summaries_path, templateType=TEMPLATE_TYPES[template_type]["template_type"])
        else:
            self.template = self.mso.query_objs(self.summaries_path)

        # Remove unwanted keys from existing object for better output and diff compares
        if isinstance(self.template, dict):
            for key in ["_updateVersion", "version"]:
                self.template.pop(key, None)

    def _set_schema_properties(self):
        self.schema_name = self.template_summary.get("schemaName")
        self.schema_id = self.template_summary.get("schemaId")
        self.schema_path = "schemas/{0}".format(self.schema_id)

    def _set_deployment_properties(self):
        # "deploySummmary" is a typo in the API
        # Checking both keys to be safe
        deploy_summary = self.template_summary.get("deploySummmary", {}) or self.template_summary.get("deploySummary", {})
        self.deploy_task_id = deploy_summary.get("teTaskId")
        self.deployed_site_ids = [summary.get("siteId") for summary in deploy_summary.get("siteDeploymentSummaries", []) if summary.get("siteId") is not None]

    @staticmethod
    def get_object_from_list(search_list, kv_list):
        """
        Get the first matched object from a list of mso object dictionaries.
        :param search_list: Objects to search through -> List.
        :param kv_list: Key/value pairs that should match in the object. -> List[KVPair(Str, Str)]
        :return: The index and details of the object. -> Item (Named Tuple)
                 Values of provided keys of all existing objects. -> List
        """

        # Sometimes the attribute returned by api might be None
        # If search_list is None, iterating over it will throw an error
        # Thus we need to return the match of None and without existing values
        if search_list is None:
            return None, []

        def kv_match(kvs, item):
            return all((item.get(kv.key) == kv.value for kv in kvs))

        match = next((Item(index, item) for index, item in enumerate(search_list) if kv_match(kv_list, item)), None)
        existing = [item.get(kv.key) for item in search_list for kv in kv_list if item.get(kv.key) is not None]
        return match, existing

    def validate_template(self, template_type):
        """
        Validate that attributes are set to a value that is not equal None.
        :return: None
        """
        if not self.template or not isinstance(self.template, dict):
            self.mso.fail_json(msg="Template '{0}' not found.".format(self.template_name))
        if self.template.get("templateType") != template_type:
            self.mso.fail_json(msg="Template type must be '{0}'.".format(template_type))

    def get_object_by_key_value_pairs(self, object_description, search_list, kv_list, fail_module=False):
        """
        Get the object from a list of mso object dictionaries by name.
        :param object_description: Description of the object to search for -> Str
        :param search_list: Objects to search through -> List.
        :param kv_list: Key/value pairs that should match in the object. -> List[KVPair(Str, Str)]
        :param fail_module: When match is not found fail the ansible module. -> Bool
        :return: The object. -> Dict | None
        """
        match, existing = self.get_object_from_list(search_list, kv_list)
        if not match and fail_module:
            msg = "Provided {0} with '{1}' not matching existing object(s): {2}".format(object_description, kv_list, ", ".join(existing))
            self.mso.fail_json(msg=msg)
        return match

    def get_object_by_uuid(self, object_description, search_list, uuid, fail_module=False):
        """
        Get the object from a list of mso object dictionaries by uuid.
        :param object_description: Description of the object to search for -> Str
        :param search_list: Objects to search through -> List.
        :param uuid: UUID of the object to search for -> Str
        :param fail_module: When match is not found fail the ansible module. -> Bool
        :return: The object. -> Dict | None
        """
        kv_list = [KVPair("uuid", uuid)]
        return self.get_object_by_key_value_pairs(object_description, search_list, kv_list, fail_module)

    def get_vlan_pool_uuid(self, vlan_pool_name):
        """
        Get the UUID of a VLAN pool by name.
        :param vlan_pool_name: Name of the VLAN pool to search for -> Str
        :return: UUID of the VLAN pool. -> Str
        """
        existing_vlan_pools = self.template.get("fabricPolicyTemplate", {}).get("template", {}).get("vlanPools", [])
        kv_list = [KVPair("name", vlan_pool_name)]
        match = self.get_object_by_key_value_pairs("VLAN Pool", existing_vlan_pools, kv_list, fail_module=True)
        return match.details.get("uuid")

    def get_vlan_pool_name(self, vlan_pool_uuid):
        """
        Get the UUID of a VLAN pool by name.
        :param vlan_pool_name: Name of the VLAN pool to search for -> Str
        :return: UUID of the VLAN pool. -> Str
        """
        existing_vlan_pools = self.template.get("fabricPolicyTemplate", {}).get("template", {}).get("vlanPools", [])
        kv_list = [KVPair("uuid", vlan_pool_uuid)]
        match = self.get_object_by_key_value_pairs("VLAN Pool", existing_vlan_pools, kv_list, fail_module=True)
        return match.details.get("name")

    def get_route_map(self, attr_name, tenant_id, tenant_name, route_map, route_map_objects):
        """
        Retrieves the details of a specific route map object based on the provided attributes.
        :param attr_name: The attribute name for error messaging. -> Str
        :param tenant_id: The ID of the tenant. -> Str
        :param tenant_name: The name of the tenant. -> Str
        :param route_map: The name of the route map. -> Str
        :param route_map_objects: The list of route map objects to search from. -> List
        :return: The details of the route map object if found, otherwise an empty dictionary. -> Dict
        """
        if route_map and tenant_id and route_map_objects:
            route_map_object = self.get_object_from_list(
                route_map_objects,
                [KVPair("name", route_map), KVPair("tenantId", tenant_id)],
            )
            if route_map_object[0]:
                return route_map_object[0].details
            else:
                self.mso.fail_json(msg="Provided Route Map {0}: {1} with the tenant: {2} not found.".format(attr_name, route_map, tenant_name))
        else:
            return {}

    def get_vrf_object(self, vrf_dict, tenant_id, templates_objects_path):
        """
        Get VRF object based on provided parameters.
        :param vrf_dict: Dictionary containing VRF details. -> Dict
        :param tenant_id: Id of the tenant. -> Str
        :param templates_objects_path: Path to the templates objects. -> Str
        :return: VRF object if found, otherwise fail with an error message. -> Dict
        """

        vrf_path = generate_api_endpoint(templates_objects_path, **{"type": "vrf", "tenant-id": tenant_id, "include-common": "true"})
        vrf_objects = self.mso.query_objs(vrf_path)
        vrf_kv_list = [
            KVPair("name", vrf_dict.get("name")),
            KVPair("templateId", vrf_dict.get("template_id")) if vrf_dict.get("template_id") else KVPair("templateName", vrf_dict.get("template")),
            KVPair("schemaId", vrf_dict.get("schema_id")) if vrf_dict.get("schema_id") else KVPair("schemaName", vrf_dict.get("schema")),
            KVPair("tenantId", tenant_id),
        ]

        vrf_object = self.get_object_from_list(vrf_objects, vrf_kv_list)

        if vrf_object[0]:
            return vrf_object[0]
        else:
            self.mso.fail_json(msg="Provided VRF {0} not found.".format(vrf_dict.get("name")))

    def get_bd_object(self, bd_dict, tenant_id, templates_objects_path):
        """
        Get Bridge Domain object based on provided parameters.
        :param bd_dict: Dictionary containing Bridge Domain details. -> Dict
        :param tenant_id: Id of the tenant. -> Str
        :param templates_objects_path: Path to the templates objects. -> Str
        :return: Bridge Domain object if found, otherwise fail with an error message. -> Dict
        """

        bd_path = generate_api_endpoint(templates_objects_path, **{"type": "bd", "tenant-id": tenant_id, "include-common": "true"})
        bd_objects = self.mso.query_objs(bd_path)
        bd_kv_list = [
            KVPair("name", bd_dict.get("name")),
            KVPair("templateId", bd_dict.get("template_id")) if bd_dict.get("template_id") else KVPair("templateName", bd_dict.get("template")),
            KVPair("schemaId", bd_dict.get("schema_id")) if bd_dict.get("schema_id") else KVPair("schemaName", bd_dict.get("schema")),
            KVPair("tenantId", tenant_id),
        ]
        return self.get_object_by_key_value_pairs("Bridge Domain", bd_objects, bd_kv_list, fail_module=True)

    def get_l3out_node_routing_policy_object(self, uuid=None, name=None, fail_module=False):
        """
        Get the L3Out Node Routing Policy by UUID or Name.
        :param uuid: UUID of the L3Out Node Routing Policy to search for -> Str
        :param name: Name of the L3Out Node Routing Policy to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        existing_l3out_node_routing_policy = self.template.get("tenantPolicyTemplate", {}).get("template", {}).get("l3OutNodePolGroups", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs(
                "L3Out Node Routing Policy", existing_l3out_node_routing_policy, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module
            )
        return existing_l3out_node_routing_policy  # Query all objects

    def get_interface_policy_group_uuid(self, interface_policy_group):
        """
        Get the UUID of an Interface Policy Group by name.
        :param interface_policy_group: Name of the Interface Policy Group to search for -> Str
        :return: UUID of the Interface Policy Group. -> Str
        """
        existing_policy_groups = self.template.get("fabricPolicyTemplate", {}).get("template", {}).get("interfacePolicyGroups", [])
        kv_list = [KVPair("name", interface_policy_group)]
        match = self.get_object_by_key_value_pairs("Interface Policy Groups", existing_policy_groups, kv_list, fail_module=True)
        return match.details.get("uuid")

    def get_ipsla_monitoring_policy(self, uuid=None, name=None, fail_module=False):
        """
        Get the IPSLA Monitoring Policy by UUID or Name.
        :param uuid: UUID of the IPSLA Monitoring Policy to search for -> Str
        :param name: Name of the IPSLA Monitoring Policy to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        existing_ipsla_policies = self.template.get("tenantPolicyTemplate", {}).get("template", {}).get("ipslaMonitoringPolicies", [])
        if name or uuid:
            return self.get_object_by_key_value_pairs(
                "IPSLA Monitoring Policy",
                existing_ipsla_policies,
                [KVPair("uuid", uuid) if uuid else KVPair("name", name)],
                fail_module=fail_module,
            )
        return existing_ipsla_policies

    def get_l3out_object(self, uuid=None, name=None, fail_module=False, search_object=None):
        """
        Get the L3Out by uuid or name.
        :param uuid: UUID of the L3Out to search for -> Str
        :param name: Name of the L3Out to search for -> Str
        :param search_object: The object to search in -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        if not search_object:
            search_object = self.template
        existing_l3outs = search_object.get("l3outTemplate", {}).get("l3outs", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs("L3Out", existing_l3outs, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module)
        return existing_l3outs  # Query all objects

    def get_l3out_node_group(self, name, l3out_object, fail_module=False):
        """
        Get the L3Out Node Group Policy by name.
        :param name: Name of the L3Out Node Group Policy to search for -> Str
        :param l3out_object: L3Out object to search Node Group Policy -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the Name is existing in the search list -> Dict
                 When the Name is not existing in the search list -> None
                 When the Name is None, and the search list is not empty -> List[Dict]
                 When the Name is None, and the search list is empty -> List[]
        """
        existing_l3out_node_groups = l3out_object.get("nodeGroups", [])
        if name:  # Query a specific object
            return self.get_object_by_key_value_pairs("L3Out Node Group Policy", existing_l3out_node_groups, [KVPair("name", name)], fail_module)
        return existing_l3out_node_groups  # Query all objects

    def get_port_channel_match(self, port_channel, mso_templates):
        """
        Get the port channel from the provided port channel reference or uuid.
        :param port_channel: The port channel object containing reference or uuid to search for -> ndo_l3out_port_channel_spec
        :param mso_templates: MSO Templates object to search for referenced templates -> MSOTemplates
        :return: The matched port channel object or None if not found -> Dict | None
        """
        port_channel_match = None
        if port_channel:
            port_channel_uuid = port_channel.get("uuid")
            if port_channel_uuid:
                port_channel_match = self.get_template_object_by_uuid("portChannel", port_channel_uuid, True)
            else:
                fabric_resource_mso_template = mso_templates.get_template(
                    "fabric_resource",
                    port_channel.get("reference").get("template"),
                    port_channel.get("reference").get("template_id"),
                    fail_module=True,
                )
                port_channel_match = fabric_resource_mso_template.get_port_channel(
                    None,
                    port_channel.get("reference").get("name"),
                    fail_module=True,
                ).details
        return port_channel_match

    def get_port_channel(self, uuid=None, name=None, fail_module=False):
        """
        Get the port channel by uuid or name.
        :param uuid: UUID of the Port Channel to search for -> Str
        :param name: Name of the Port Channel to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        existing_port_channels = self.template.get("fabricResourceTemplate", {}).get("template", {}).get("portChannels", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs(
                "Port Channel", existing_port_channels, [KVPair("uuid", uuid)] if uuid else [KVPair("name", name)], fail_module=fail_module
            )
        return existing_port_channels

    def get_virtual_port_channel_match(self, virtual_port_channel, mso_templates):
        """
        Get the virtual port channel from the provided virtual port channel reference or uuid.
        :param virtual_port_channel: The virtual port channel object containing reference or uuid to search for -> ndo_l3out_virtual_port_channel_spec
        :param mso_templates: MSO Templates object to search for referenced templates -> MSOTemplates
        :return: The matched virtual port channel object or None if not found -> Dict | None
        """
        virtual_port_channel_match = None
        if virtual_port_channel:
            virtual_port_channel_uuid = virtual_port_channel.get("uuid")
            if virtual_port_channel_uuid:
                virtual_port_channel_match = self.get_template_object_by_uuid("virtualPortChannel", virtual_port_channel_uuid, True)
            else:
                fabric_resource_mso_template = mso_templates.get_template(
                    "fabric_resource",
                    virtual_port_channel.get("reference").get("template"),
                    virtual_port_channel.get("reference").get("template_id"),
                    fail_module=True,
                )
                virtual_port_channel_match = fabric_resource_mso_template.get_virtual_port_channel(
                    virtual_port_channel_uuid,
                    virtual_port_channel.get("reference").get("name"),
                    fail_module=True,
                ).details
        return virtual_port_channel_match

    def get_virtual_port_channel(self, uuid=None, name=None, fail_module=False):
        """
        Get the virtual port channel by uuid or name.
        :param uuid: UUID of the Virtual Port Channel to search for -> Str
        :param name: Name of the Virtual Port Channel to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        existing_virtual_port_channels = self.template.get("fabricResourceTemplate", {}).get("template", {}).get("virtualPortChannels", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs(
                "Virtual Port Channel", existing_virtual_port_channels, [KVPair("uuid", uuid)] if uuid else [KVPair("name", name)], fail_module=fail_module
            )
        return existing_virtual_port_channels

    def get_l3out_node(self, l3out_object, pod_id, node_id, fail_module=False):
        """
        Get the L3Out Node by pod_id and node_id.
        :param l3out_object: L3Out object to search for the Node -> Dict
        :param pod_id: Pod ID of the Node to search for -> Str
        :param node_id: Node ID of the Node to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the pod_id | node_id is existing in the search list -> Dict
                 When the pod_id | node_id is not existing in the search list -> None
                 When both pod_id and node_id are None, and the search list is not empty -> List[Dict]
                 When both pod_id and node_id are None, and the search list is empty -> List[]
        """
        existing_l3out_nodes = l3out_object.get("nodes", [])
        if pod_id and node_id:  # Query a specific object
            return self.get_object_by_key_value_pairs("L3Out Node", existing_l3out_nodes, [KVPair("podID", pod_id), KVPair("nodeID", node_id)], fail_module)
        return existing_l3out_nodes  # Query all objects

    def get_l3out_node_static_route(self, node_object, prefix, fail_module=False):
        """
        Get the L3Out Node Static Route by prefix.
        :param node_object: L3Out Node object to search for the Static Route -> Dict
        :param prefix: Prefix of the Static Route to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the prefix is existing in the search list -> Dict
                 When the prefix is not existing in the search list -> None
                 When the prefix is None, and the search list is not empty -> List[Dict]
                 When the prefix is None, and the search list is empty -> List[]
        """
        existing_l3out_static_routes = node_object.get("staticRoutes", [])
        if prefix:  # Query a specific object
            return self.get_object_by_key_value_pairs("L3Out Node Static Route", existing_l3out_static_routes, [KVPair("prefix", prefix)], fail_module)
        return existing_l3out_static_routes  # Query all objects

    def get_l3out_node_static_route_next_hop(self, static_route_object, ip, fail_module=False):
        """
        Get the L3Out Node Static Route Next Hop by IP.
        :param static_route_object: L3Out Node Static Route object to search for the Next Hop -> Dict
        :param ip: IP of the Static Route Next Hop to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the IP is existing in the search list -> Dict
                 When the IP is not existing in the search list -> None
                 When the IP is None, and the search list is not empty -> List[Dict]
                 When the IP is None, and the search list is empty -> List[]
        """
        existing_l3out_static_route_next_hops = static_route_object.get("nextHops", [])
        if ip:  # Query a specific object
            return self.get_object_by_key_value_pairs(
                "L3Out Node Static Route Next Hop", existing_l3out_static_route_next_hops, [KVPair("nextHopIP", ip)], fail_module
            )
        return existing_l3out_static_route_next_hops  # Query all objects

    def get_ipsla_track_list(self, uuid=None, name=None, fail_module=False):
        """
        Get the IPSLA Track List by uuid or name.
        :param uuid: UUID of the IPSLA Track List to search for -> Str
        :param name: Name of the IPSLA Track List to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        existing_ipsla_track_lists = self.template.get("tenantPolicyTemplate", {}).get("template", {}).get("ipslaTrackLists", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs(
                "IPSLA Track List",
                existing_ipsla_track_lists,
                [KVPair("uuid", uuid)] if uuid else [KVPair("name", name)],
                fail_module=fail_module,
            )
        return existing_ipsla_track_lists  # Query all objects

    def get_l3out_secondary_address(self, parent_object, parent_type, secondary_address, side_b, fail_module=False):
        """
        Get the L3Out Secondary Address by address.
        :param parent_object: The parent object to search for the secondary IP address -> Dict
        :param secondary_address: The secondary address to search for -> Str
        :param side_b: The side indicator for the SVI VPC parent object.
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the address is existing in the search list -> Dict
                 When the address is not existing in the search list -> None
                 When the address is None, and the search list is not empty -> List[Dict]
                 When the address is None, and the search list is empty -> List[]
        """
        if parent_type == "floating_svi_path_attributes":
            existing_secondary_address = parent_object.get("secondaryAddresses", [])
        else:
            existing_secondary_address = parent_object.get("sideBAddresses" if side_b else "addresses", {}).get("secondary", [])

        if secondary_address:  # Query a specific object
            kv_list = [KVPair("address", secondary_address)]
            return self.get_object_by_key_value_pairs("L3Out Secondary IP Address", existing_secondary_address, kv_list, fail_module)
        return existing_secondary_address  # Query all objects

    def get_l3out_routed_interface(self, l3out_object, pod_id, node_id, path, path_ref, fail_module=False):
        """
        Get the L3Out Routed Interface by pod_id, node_id, path, and path_ref.
        :param l3out_object: L3Out object to search for the Routed Interface -> Dict
        :param pod_id: Pod ID of the Routed Interface to search for -> Str
        :param node_id: Node ID of the Routed Interface to search for -> Str
        :param path: Path of the Routed Interface to search for -> Str
        :param path_ref: Path reference of the Routed Interface to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the pod_id, node_id, path | path_ref is existing in the search list -> Dict
                 When the pod_id, node_id, path | path_ref is not existing in the search list -> None
                 When both pod_id, node_id, path and path_ref are None, and the search list is not empty -> List[Dict]
                 When both pod_id, node_id, path and path_ref are None, and the search list is empty -> List[]
        """
        existing_l3out_interfaces = l3out_object.get("interfaces", [])
        if (pod_id and node_id and path) or path_ref:  # Query a specific object
            if path_ref:
                kv_list = [KVPair("pathRef", path_ref)]
            else:
                kv_list = [KVPair("podID", pod_id), KVPair("nodeID", node_id), KVPair("path", path)]

            return self.get_object_by_key_value_pairs("L3Out Interface", existing_l3out_interfaces, kv_list, fail_module)
        return existing_l3out_interfaces  # Query all objects

    def get_l3out_routed_sub_interface(self, l3out_object, pod_id, node_id, path, path_ref, encap, fail_module=False):
        """
        Get the L3Out Routed Sub-Interface by pod_id, node_id, path, and path_ref.
        :param l3out_object: L3Out object to search for the Routed Sub-Interface -> Dict
        :param pod_id: Pod ID of the Routed Sub-Interface to search for -> Str
        :param node_id: Node ID of the Routed Sub-Interface to search for -> Str
        :param path: Path of the Routed Sub-Interface to search for -> Str
        :param path_ref: Path reference of the Routed Sub-Interface to search for -> Str
        :param encap: Encapsulation details of the Routed Sub-Interface to search for -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the pod_id, node_id, path | path_ref with encap is existing in the search list -> Dict
                 When the pod_id, node_id, path | path_ref with encap is not existing in the search list -> None
                 When both pod_id, node_id, path, path_ref, and encap are None, and the search list is not empty -> List[Dict]
                 When both pod_id, node_id, path, path_ref, and encap are None, and the search list is empty -> List[]
        """
        existing_l3out_interfaces = l3out_object.get("subInterfaces", [])
        if ((pod_id and node_id and path) or path_ref) and encap:  # Query a specific object
            if path_ref:
                kv_list = [KVPair("pathRef", path_ref), KVPair("encap", encap)]
            else:
                kv_list = [KVPair("podID", pod_id), KVPair("nodeID", node_id), KVPair("path", path), KVPair("encap", encap)]

            return self.get_object_by_key_value_pairs("L3Out Sub-Interface", existing_l3out_interfaces, kv_list, fail_module)
        return existing_l3out_interfaces  # Query all objects

    def get_l3out_svi_interface(self, l3out_object, pod_id, node_id, path, encap, path_ref, fail_module=False):
        """
        Get the L3Out SVI Interface by pod_id, node_id, path, and path_ref.
        :param l3out_object: L3Out object to search for the SVI Interface -> Dict
        :param pod_id: Pod ID of the SVI Interface to search for -> Str
        :param node_id: Node ID of the SVI Interface to search for -> Str
        :param path: Path of the SVI Interface to search for -> Str
        :param encap: Encapsulation details of the Floating SVI Interface to search for -> Dict
        :param path_ref: Path reference of the SVI Interface to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the pod_id, node_id, path | path_ref is existing in the search list -> Dict
                 When the pod_id, node_id, path | path_ref is not existing in the search list -> None
                 When both pod_id, node_id, path and path_ref are None, and the search list is not empty -> List[Dict]
                 When both pod_id, node_id, path and path_ref are None, and the search list is empty -> List[]
        """
        existing_l3out_svi_interfaces = l3out_object.get("sviInterfaces", [])
        if encap and ((pod_id and node_id and path) or path_ref):  # Query a specific object
            if path_ref:
                kv_list = [KVPair("pathRef", path_ref), KVPair("encap", encap)]
            else:
                kv_list = [KVPair("podID", pod_id), KVPair("nodeID", node_id), KVPair("path", path), KVPair("encap", encap)]

            return self.get_object_by_key_value_pairs("L3Out SVI Interface", existing_l3out_svi_interfaces, kv_list, fail_module)
        return existing_l3out_svi_interfaces  # Query all objects

    def get_l3out_floating_svi_interface(self, l3out_object, pod_id, node_id, encap, fail_module=False):
        """
        Get the L3Out Floating SVI Interface by pod_id and node_id.
        :param l3out_object: L3Out object to search for the Floating SVI Interface -> Dict
        :param pod_id: Pod ID of the Floating SVI Interface to search for -> Str
        :param node_id: Node ID of the Floating SVI Interface to search for -> Str
        :param encap: Encapsulation details of the Floating SVI Interface to search for -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the pod_id, node_id and encap are existing in the search list -> Dict
                 When the pod_id, node_id and encap are not existing in the search list -> None
                 When both pod_id, node_id, and encap are None, and the search list is not empty -> List[Dict]
                 When both pod_id, node_id, and encap are None, and the search list is empty -> List[]
        """
        existing_l3out_floating_svi_interfaces = l3out_object.get("floatingSviInterfaces", [])
        if pod_id and node_id and encap:  # Query a specific object
            kv_list = [KVPair("podID", pod_id), KVPair("nodeID", node_id), KVPair("encap", encap)]

            return self.get_object_by_key_value_pairs("L3Out Floating SVI Interface", existing_l3out_floating_svi_interfaces, kv_list, fail_module)
        return existing_l3out_floating_svi_interfaces  # Query all objects

    def get_l3out_floating_svi_interface_path_attributes(self, l3out_floating_svi_interface_object, domain_type, domain, fail_module=False):
        """
        Get the L3Out Floating SVI Interface Path Attributes by domain_type and domain.
        :param l3out_floating_svi_interface_object: L3Out Floating SVI Interface object to search for the Path Attributes -> Dict
        :param domain_type: Domain type of the Path Attributes to search for -> Str
        :param domain: Domain of the Path Attributes to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the domain_type and domain are existing in the search list -> Dict
                 When the domain_type and domain are not existing in the search list -> None
                 When both domain_type and domain are None, and the search list is not empty -> List[Dict]
                 When both domain_type and domain are None, and the search list is empty -> List[]
        """
        existing_path_attributes = l3out_floating_svi_interface_object.get("svi", {}).get("floatingPathAttributes", [])
        if domain_type and domain:  # Query a specific object
            kv_list = [KVPair("domainType", domain_type), KVPair("domain", domain)]
            return self.get_object_by_key_value_pairs("L3Out Floating SVI Interface Path Attributes", existing_path_attributes, kv_list, fail_module)
        return existing_path_attributes  # Query all objects

    def get_node_settings_object(self, uuid=None, name=None, fail_module=False):
        """
        Get the Fabric Node Settings by uuid or name.
        :param uuid: UUID of the Node Setting to search for -> Str
        :param name: Name of the Node Setting to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        existing_objects = self.template.get("fabricPolicyTemplate", {}).get("template", {}).get("nodePolicyGroups", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs("Node Settings", existing_objects, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module)
        return existing_objects  # Query all objects

    def get_pod_profile_object(self, uuid=None, name=None, search_object=None, fail_module=False):
        """
        Get the Pod Profile by uuid or name.
        :param uuid: UUID of the Pod Profile to search for -> Str
        :param name: Name of the Pod Profile to search for -> Str
        :param search_object: The object to search in -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        if not search_object:
            search_object = self.template
        existing_objects = search_object.get("fabricResourceTemplate", {}).get("template", {}).get("podProfiles", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs("Pod Profile", existing_objects, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module)
        return existing_objects  # Query all objects

    def get_pod_settings_object(self, uuid=None, name=None, search_object=None, fail_module=False):
        """
        Get the Pod Settings by uuid or name.
        :param uuid: UUID of the Pod Settings to search for -> Str
        :param name: Name of the Pod Settings to search for -> Str
        :param search_object: The object to search in -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        if not search_object:
            search_object = self.template
        existing_objects = search_object.get("fabricPolicyTemplate", {}).get("template", {}).get("podPolicyGroups", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs("Pod Settings", existing_objects, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module)
        return existing_objects  # Query all objects

    def get_ntp_policy_object(self, uuid=None, name=None, search_object=None, fail_module=False):
        """
        Get the NTP Policy by uuid or name.
        :param uuid: UUID of the NTP Policy to search for -> Str
        :param name: Name of the NTP Policy to search for -> Str
        :param search_object: The object to search in -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        if not search_object:
            search_object = self.template
        existing_objects = search_object.get("fabricPolicyTemplate", {}).get("template", {}).get("ntpPolicies", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs("NTP Policy", existing_objects, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module)
        return existing_objects  # Query all objects

    def get_ptp_policy_profile_object(self, uuid=None, name=None, fail_module=False):
        # This object is an exception where the template can only contain a single PTP Policy, thus an additional layer of nested query is done
        """
        Get the PTP Policy Profile by UUID or Name.
        :param uuid: UUID of the PTP Profile to search for -> Str
        :param name: Name of the PTP Profile to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        existing_ptp_profile = self.template.get("fabricPolicyTemplate", {}).get("template", {}).get("ptpPolicy", {}).get("profiles", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs(
                "PTP Policy Profile", existing_ptp_profile, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module
            )
        return existing_ptp_profile  # Query all objects

    def get_macsec_policy_object(self, uuid=None, name=None, search_object=None, fail_module=False):
        """
        Get the MACsec Policy by uuid or name.
        :param uuid: UUID of the MACsec Policy to search for -> Str
        :param name: Name of the MACsec Policy to search for -> Str
        :param search_object: The object to search in -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        if not search_object:
            search_object = self.template
        existing_objects = search_object.get("fabricPolicyTemplate", {}).get("template", {}).get("macsecPolicies", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs("MACsec Policy", existing_objects, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module)
        return existing_objects  # Query all objects

    def get_l3out_interface_routing_policy_object(self, uuid=None, name=None, fail_module=False):
        """
        Get the L3Out Interface Routing Policy by UUID or Name.
        :param uuid: UUID of the L3Out Interface Routing Policy to search for -> Str
        :param name: Name of the L3Out Interface Routing Policy to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        existing_l3out_interface_routing_policy = self.template.get("tenantPolicyTemplate", {}).get("template", {}).get("l3OutIntfPolGroups", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs(
                "L3Out Interface Routing Policy",
                existing_l3out_interface_routing_policy,
                [KVPair("uuid", uuid) if uuid else KVPair("name", name)],
                fail_module,
            )
        return existing_l3out_interface_routing_policy  # Query all objects

    def get_match_rule_policy_object(self, uuid=None, name=None, search_object=None, fail_module=False):
        """
        Get the Match Rule Policy by uuid or name.
        :param uuid: UUID of the Match Rule Policy to search for -> Str
        :param name: Name of the Match Rule Policy to search for -> Str
        :param search_object: The object to search in -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        if not search_object:
            search_object = self.template
        existing_objects = search_object.get("tenantPolicyTemplate", {}).get("template", {}).get("matchRulePolicies", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs(
                "Match Rule Policy",
                existing_objects,
                [KVPair("uuid", uuid) if uuid else KVPair("name", name)],
                fail_module,
            )
        return existing_objects  # Query all objects

    def get_endpoint_mac_tag_policy_object(self, uuid=None, mac=None, bd_uuid=None, vrf_uuid=None, search_object=None, fail_module=False):
        """
        Get the Endpoint MAC Tag Policy by uuid or mac and bd_uuid or vrf_uuid.
        :param uuid: UUID of the Endpoint MAC Tag Policy to search for -> Str
        :param mac: The MAC address of the Endpoint MAC Tag Policy to search for -> Str
        :param bd_uuid: UUID of the Bridge Domain referenced by the Endpoint MAC Tag Policy to search for -> Str
        :param vrf_uuid: UUID of the VRF referenced by the Endpoint MAC Tag Policy to search for -> Str
        :param search_object: The object to search in -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | MAC address and (VRF UUID | Bridge Domain UUID) are existing in the search list -> Dict
                 When the UUID | MAC address and (VRF UUID | Bridge Domain UUID) are not existing in the search list -> None
                 When UUID is None and MAC address | (VRF UUID and Bridge Domain UUID) are None, and the search list is not empty -> List[Dict]
                 When UUID and MAC address | (VRF UUID and Bridge Domain UUID) are None, and the search list is empty -> List[]
        """
        kv_list = []
        if not search_object:
            search_object = self.template
        existing_objects = search_object.get("tenantPolicyTemplate", {}).get("template", {}).get("endpointMacTagPolicies", [])
        if uuid:
            kv_list = [KVPair("uuid", uuid)]
        elif mac and bd_uuid:
            kv_list = [KVPair("mac", mac), KVPair("bdRef", bd_uuid)]
        elif mac and vrf_uuid:
            kv_list = [KVPair("mac", mac), KVPair("vrfRef", vrf_uuid)]
        if kv_list:
            return self.get_object_by_key_value_pairs("Endpoint MAC Tag Policy", existing_objects, kv_list, fail_module)
        return existing_objects  # Query all objects

    def get_endpoint_ip_tag_policy_object(self, uuid=None, ip=None, vrf_uuid=None, search_object=None, fail_module=False):
        """
        Get the Endpoint IP Tag Policy by uuid or ip and vrf_uuid.
        :param uuid: UUID of the Endpoint IP Tag Policy to search for -> Str
        :param ip: The IP address of the Endpoint IP Tag Policy to search for -> Str
        :param vrf_uuid: UUID of the VRF referenced by the Endpoint IP Tag Policy to search for -> Str
        :param search_object: The object to search in -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                When the UUID | IP address and VRF UUID are existing in the search list -> Dict
                When the UUID | IP address and VRF UUID are not existing in the search list -> None
                When UUID is None and IP address | VRF UUID is None, and the search list is not empty -> List[Dict]
                When UUID is None and IP address | VRF UUID is None, and the search list is empty -> List[]
        """
        if not search_object:
            search_object = self.template
        existing_objects = search_object.get("tenantPolicyTemplate", {}).get("template", {}).get("endpointIPTagPolicies", [])
        if uuid or (ip and vrf_uuid):
            if uuid:
                kv_list = [KVPair("uuid", uuid)]
            else:
                kv_list = [KVPair("ip", ip), KVPair("vrfRef", vrf_uuid)]
            return self.get_object_by_key_value_pairs("Endpoint IP Tag Policy", existing_objects, kv_list, fail_module)
        return existing_objects  # Query all objects

    def get_direct_child_object(self, parent_object, description, endpoint, identifiers=None, fail_module=False):
        """
        Get the direct child object using its identifiers and its parent object.
        :param parent_object: Parent object data where to search the direct child object -> Dict
        :param description: Description of the child object to search for -> Str
        :param endpoint: NDO API child object's endpoint -> Str
        :param identifiers: child object's identifiers with coresponding identifier's name and value -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When one of the child object identifiers is existing in the search list -> Dict
                 When the child object identifiers are not existing in the search list -> None
                 When the child object identifiers are None, and the search list is not empty -> List[Dict]
                 When the child object identifiers ae None, and the search list is empty -> List[]
        """
        if isinstance(identifiers, dict) and identifiers.values():  # Query a specific object
            for key, value in identifiers.items():
                if value:
                    child_object_kvpair = KVPair(key, value)
            return self.get_object_by_key_value_pairs(
                description,
                parent_object.details.get(endpoint, []),
                [child_object_kvpair],
                fail_module,
            )
        return parent_object.details.get(endpoint, [])  # Query all objects

    def get_template_policy_uuid(self, template_type, policy_name, policy_type):
        """
        Get the UUID of a Tenant Policy by name.
        :param template_type: The type of template -> Str
        :param policy_name: Name of the policy -> Str
        :param policy_type: The type of the policy specified in the API response -> Str
        :return: UUID of the tenant policy -> Str
        """
        existing_policies = self.template.get(TEMPLATE_TYPES[template_type]["template_type_container"], {}).get("template", {}).get(policy_type, [])
        match = self.get_object_by_key_value_pairs(policy_type, existing_policies, [KVPair("name", policy_name)], fail_module=True)
        return match.details.get("uuid")

    def clear_template_objects_cache(self):
        self.template_objects_cache = {}

    def get_template_object_name_by_uuid(self, object_type, uuid, fail_module=True):
        """
        Retrieve the name of a specific object type in the MSO template using its UUID.
        :param mso: An instance of the MSO class, which provides methods for making API requests -> MSO Class instance
        :param object_type: The type of the object to retrieve the name for -> Str
        :param uuid: The UUID of the object to retrieve the name for -> Str
        :return: Str | None: The processed result which could be:
              When the UUID is existing, returns object name -> Str
              When the UUID is not existing -> None
        """
        response_object = self.get_template_object_by_uuid(object_type, uuid, fail_module)
        if response_object:
            return response_object.get("name")

    def get_template_object_by_uuid(self, object_type, uuid, fail_module=True, use_cache=False):
        """
        Retrieve a specific object type in the MSO template using its UUID.
        :param object_type: The type of the object to retrieve -> Str
        :param uuid: The UUID of the object to retrieve -> Str
        :param use_cache: Use the cached result of the templates/objects API for the UUID -> Bool
        :return: Dict | None: The processed result which could be:
            When the UUID is existing, returns object -> Dict
            When the UUID is not existing -> None
        """
        response_object = None
        if use_cache and uuid in self.template_objects_cache.keys():
            response_object = self.template_objects_cache[uuid]
        else:
            response_object = self.mso.request("templates/objects?type={0}&uuid={1}".format(object_type, uuid), "GET")
            self.template_objects_cache[uuid] = response_object
        if not response_object and fail_module:
            msg = "Provided {0} with UUID of '{1}' not found.".format(object_type, uuid)
            self.mso.fail_json(msg=msg)
        return response_object

    def update_config_with_template_and_references(self, config_data, reference_collections=None, set_template=True, use_cache=False):
        """
        Return the updated config_data with the template values and reference_collections if provided
        :param config_data: The original config_data that requires to be updated -> Dict
        :param reference_collections: A dict containing the object type, references and the corresponding names -> Dict
        :param set_template: Adds the templateId and templateName to the config_data -> Bool
        :param use_cache: Use the cached result of the templates/objects API for the ref UUID -> Bool
        :return: Updated config_data with names for references -> Dict
        Example 1:
        reference_collections = {
            "qos": {
                "name": "qosName",
                "reference": "qosRef",
                "type": "qos",
                "template": "qosTemplateName",
                "templateId": "qosTemplateId",
            },
            "interfaceRoutingPolicy": {
                "name": "interfaceRoutingPolicyName",
                "reference": "interfaceRoutingPolicyRef",
                "type": "l3OutIntfPolGroup",
                "template": "interfaceRoutingPolicyTemplateName",
                "templateId": "interfaceRoutingPolicyTemplateId",
            },
        }
        config_data = {
            "qosRef": "unique-qos-id",
            "interfaceRoutingPolicyRef": "unique-interface-id"
        }
        updated_config_data = mso_template_object.update_config_with_template_and_references(mso_instance, config_data, reference_collections)
        Expected Output:
        {    "templateName": "template_name",
             "templateId": "unique-template-id",
             "qosRef": "unique-qos-id",
             "interfaceRoutingPolicyRef": "unique-interface-id",
             "qosName": "Resolved QoS Name",
             "qosTemplateName": "Resolved QoS Template Name",
             "qosTemplateId": "Resolved QoS Template ID",
             "interfaceRoutingPolicyName": "Resolved Interface Routing Policy Name",
             "interfaceRoutingPolicyTemplateName": "Resolved Interface Routing Policy Template Name",
             "interfaceRoutingPolicyTemplateId": "Resolved Interface Routing Policy Template ID"
         }
        Example 2:
        reference_collections = {
            "stateLimitRouteMap": {
                "name": "stateLimitRouteMapName",
                "reference": "stateLimitRouteMapRef",
                "type": "mcastRouteMap"
            },
            "reportPolicyRouteMap": {
                "name": "reportPolicyRouteMapName",
                "reference": "reportPolicyRouteMapRef",
                "type": "mcastRouteMap"
            },
            "staticReportRouteMap": {
                "name": "staticReportRouteMapName",
                "reference": "staticReportRouteMapRef",
                "type": "mcastRouteMap"
            },
        }
        config_data = {
            "stateLimitRouteMapRef": "unique-state-limit-id",
            "reportPolicyRouteMapRef": "unique-report-policy-id"
        }
        updated_config_data = mso_template_object.update_config_with_template_and_references(mso_instance, config_data, reference_collections)
         Expected Output:
         {   "templateName": "template_name",
             "templateId": "unique-template-id",
             "stateLimitRouteMapRef": "unique-state-limit-id",
             "reportPolicyRouteMapRef": "unique-report-policy-id",
             "stateLimitRouteMapName": "Resolved State Limit Route Map Name",
             "reportPolicyRouteMapName": "Resolved Report Policy Route Map Name"
         }
        """

        # Set template ID and template name if available
        if set_template:
            if self.template_id:
                config_data["templateId"] = self.template_id
            if self.template_name:
                config_data["templateName"] = self.template_name
            if self.schema_id:
                config_data["schemaId"] = self.schema_id
            if self.schema_name:
                config_data["schemaName"] = self.schema_name

        # Update config data with reference names if reference_collections is provided
        if reference_collections:
            for reference_details in reference_collections.values():
                if config_data.get(reference_details.get("reference")):
                    template_object = self.get_template_object_by_uuid(
                        reference_details.get("type"), config_data.get(reference_details.get("reference")), True, use_cache
                    )
                    config_data[reference_details.get("name")] = template_object.get("name")
                    if reference_details.get("template"):
                        config_data[reference_details.get("template")] = template_object.get("templateName")
                    if reference_details.get("templateId"):
                        config_data[reference_details.get("templateId")] = template_object.get("templateId")
                    if reference_details.get("schemaId"):
                        config_data[reference_details.get("schemaId")] = template_object.get("schemaId")
                    if reference_details.get("schema"):
                        config_data[reference_details.get("schema")] = template_object.get("schemaName")
            for config_value in config_data.values():
                if isinstance(config_value, dict):
                    self.update_config_with_template_and_references(config_value, reference_collections, False, use_cache)
                elif isinstance(config_value, list):
                    for item in config_value:
                        if isinstance(item, dict):
                            self.update_config_with_template_and_references(item, reference_collections, False, use_cache)
        return config_data

    def update_config_with_port_channel_references(self, update_object):
        if update_object:
            reference_details = None
            if update_object.get("pathRef"):
                if update_object.get("pathType") == "vpc":
                    reference_details = {
                        "virtual_port_channel_reference": {
                            "name": "virtualPortChannelName",
                            "reference": "pathRef",
                            "type": "virtualPortChannel",
                            "template": "virtualPortChannelTemplateName",
                            "templateId": "virtualPortChannelTemplateId",
                        }
                    }
                elif update_object.get("pathType") == "pc":
                    reference_details = {
                        "port_channel_reference": {
                            "name": "portChannelName",
                            "reference": "pathRef",
                            "type": "portChannel",
                            "template": "portChannelTemplateName",
                            "templateId": "portChannelTemplateId",
                        }
                    }
            self.update_config_with_template_and_references(
                update_object,
                reference_details,
                True,
            )

    def update_config_with_node_references(self, interface, l3out_object):

        pod_id = interface.get("podID")
        node_id = interface.get("nodeID")
        node_id_2 = None

        if interface.get("pathType") == "pc":
            interface_details = self.mso.get_site_interface_details(
                self.template.get("l3outTemplate", {}).get("siteId"),
                port_channel_uuid=interface.get("pathRef"),
            )
            pod_id = interface_details.get("pod")
            node_id = interface_details.get("node")
        elif interface.get("pathType") == "vpc":
            interface_details = self.mso.get_site_interface_details(
                self.template.get("l3outTemplate", {}).get("siteId"),
                virtual_port_channel_uuid=interface.get("pathRef"),
            )
            pod_id = interface_details.get("pod")
            node_id = interface_details.get("node1")
            node_id_2 = interface_details.get("node2")

        node = self.get_l3out_node(l3out_object.details, pod_id, node_id)
        if node and not isinstance(node, list):
            interface["node"] = node.details

        if node_id_2:
            node_2 = self.get_l3out_node(l3out_object.details, pod_id, node_id_2)
            if node_2 and not isinstance(node_2, list):
                interface["sideBNode"] = node_2.details

    def update_config_with_ptp_references(self, routed_interface, mso_templates):
        ptp = routed_interface.get("ptpConfig")
        if ptp:
            reference_details = {
                "ptp_profile_reference": {
                    "name": "ptpProfileName",
                    "reference": "ptpProfileRef",
                    "type": "ptpProfile",
                    "template": "ptpProfileTemplateName",
                    "templateId": "ptpProfileTemplateId",
                }
            }
            self.update_config_with_template_and_references(
                ptp,
                reference_details,
                False,
            )
            # The PTP policy details cannot be updated via update_config_with_template_and_references because the object type is not supported.
            # The PTP policy details cannot be updated because the uuid is unknown and can only be retrieved from the template.
            # Adding additional logic to populate the parent details in output.
            ptpPolicy = (
                mso_templates.get_template(
                    "fabric_policy",
                    ptp.get("ptpProfileTemplateName"),
                    ptp.get("ptpProfileTemplateId"),
                    fail_module=True,
                )
                .template.get("fabricPolicyTemplate", {})
                .get("template", {})
                .get("ptpPolicy", {})
            )
            routed_interface["ptpConfig"]["ptpPolicyName"] = ptpPolicy.get("name")
            routed_interface["ptpConfig"]["ptpPolicyRef"] = ptpPolicy.get("uuid")

    def update_match_rule_policy_child_object_with_template_and_parent(self, match_rule_policy, config_data):
        """
        Return the updated Match Rule Policy child object config_data with the template and policy values
        :param config_data: The Match Rule Policy data -> Dict
        :param config_data: The original config_data that requires to be updated -> Dict
        :return: Updated config_data with names and ids for template and Match Rule Policy -> Dict
        """
        self.update_config_with_template_and_references(config_data)
        if match_rule_policy.get("uuid"):
            config_data["matchRulePolicyUuid"] = match_rule_policy["uuid"]
        if match_rule_policy.get("name"):
            config_data["matchRulePolicyName"] = match_rule_policy["name"]
        return config_data

    def get_route_map_policy_for_multicast_uuid(self, route_map_policy_for_multicast_name):
        """
        Get the UUID of an Route Map Policy for Multicast by name.
        :param route_map_policy_for_multicast_name: Name of the Route Map Policy for Multicast to search for -> Str
        :return: UUID of the Route Map Policy for Multicast. -> Str
        """
        existing_route_map_policies = self.template.get("tenantPolicyTemplate", {}).get("template", {}).get("mcastRouteMapPolicies", [])
        kv_list = [KVPair("name", route_map_policy_for_multicast_name)]
        match = self.get_object_by_key_value_pairs("Route Map Policy for Multicast", existing_route_map_policies, kv_list, fail_module=True)
        return match.details.get("uuid")

    def get_fabric_template_object_by_key_value(self, object_type, object_description, kv_list, fail_module=False):
        """
        Get the Fabric Policy by policy type and search criteria.
        The search criteria could be name and UUID of the object.
        :param object_type: The type of the object to retrieve the name for -> Str
        :param object_description: Description of the object to search for -> Str
        :param kv_list: Key/value pairs that should match in the object. -> List[KVPair(Str, Str)]
        :param fail_module: When match is not found fail the ansible module. -> Bool
        :return: Dict | None: The processed result which could be:
              When the object is existing in the search list -> Dict
              When the object is not existing -> None
        """
        response_object = self.mso.request("getfabricpolicies?type={0}".format(object_type), "GET")
        search_list = response_object.get("items", [{"spec": {"policies": []}}])[0].get("spec", {}).get("policies", [])
        match = self.get_object_by_key_value_pairs(object_description, search_list, kv_list, fail_module)
        if match:
            return match.details

    def get_fabric_span_session(self, uuid=None, name=None, fail_module=False):
        """
        Get the Fabric SPAN Session by uuid or name.
        :param uuid: UUID of the Fabric SPAN Session to search for -> Str
        :param name: Name of the Fabric SPAN Session to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        existing_objects = self.template.get("monitoringTemplate", {}).get("template", {}).get("spanSessions", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs("SPAN Session", existing_objects, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module)
        return existing_objects  # Query all objects

    def get_fabric_span_session_source(self, name, search_list, fail_module=False):
        """
        Get the Fabric SPAN Session Source by name.
        :param name: Name of the Fabric SPAN Session Source to search for -> Str
        :param search_list: Objects to search through -> List.
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | List[Dict] | List[]: The processed result which could be:
                 When the Name is existing in the search list -> Dict
                 When both Name is None, and the search list is not empty -> List[Dict]
                 When both Name is None, and the search list is empty -> List[]
        """
        if name and search_list:  # Query a specific object
            return self.get_object_by_key_value_pairs("SPAN Session Source", search_list, [KVPair("name", name)], fail_module)
        return search_list  # Query all objects

    def get_application_template_contract(self, uuid=None, name=None, fail_module=False):
        """
        Get the Application Template Contract by uuid or name.
        :param uuid: UUID of the Contract to search for -> Str
        :param name: Name of the Contract to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        existing_objects = self.template.get("appTemplate", {}).get("template", {}).get("contracts", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs(
                "Template Contract", existing_objects, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module
            )
        return existing_objects  # Query all objects

    def get_parent_details_for_nested_object_in_l3out(self, mso_templates, l3out_object):
        """
        Get the parent details for a nested object in the L3Out object.
        :param mso_templates: The MSO templates to search through -> List[Dict]
        :param l3out_object: The L3Out object to find the parent for -> Dict
        :return: The parent details for the nested object -> Dict, Str | None, None
        """
        port_channel_uuid = virtual_port_channel_uuid = pod_id = encap = parent_match = parent_path = None
        parent_type = self.mso.params.get("parent_type")
        node_group = self.mso.params.get("node_group")
        node_id = self.mso.params.get("node_id")
        path = self.mso.params.get("path")
        port_channel = self.get_port_channel_match(self.mso.params.get("port_channel"), mso_templates)
        if port_channel:
            port_channel_uuid = port_channel.get("uuid")
        virtual_port_channel = self.get_virtual_port_channel_match(self.mso.params.get("virtual_port_channel"), mso_templates)
        if virtual_port_channel:
            virtual_port_channel_uuid = virtual_port_channel.get("uuid")
        encapsulation_type = self.mso.params.get("encapsulation_type")
        encapsulation_value = self.mso.params.get("encapsulation_value")
        domain_type = DOMAIN_TYPE_MAP.get(self.mso.params.get("domain_type"))
        domain_provider = self.mso.params.get("domain_provider")
        if domain_type == "physicalDomain":
            domain = "uni/phys-{0}".format(self.mso.params.get("domain"))
        elif domain_type == "vmmDomain":
            domain = "uni/vmmp-{0}/dom-{1}".format(VM_DOMAIN_PROVIDER_MAP.get(self.mso.params.get("domain_provider")), self.mso.params.get("domain"))

        if node_id or path or port_channel or virtual_port_channel:
            pod_id = self.mso.get_site_interface_details(
                site_id=self.template.get("l3outTemplate", {}).get("siteId"),
                node=node_id,
                port=path,
                port_channel_uuid=port_channel_uuid,
                virtual_port_channel_uuid=virtual_port_channel_uuid,
            )
            if path or port_channel or virtual_port_channel:
                pod_id = pod_id.get("pod")

        if encapsulation_type and encapsulation_value:
            encap = {"encapType": encapsulation_type, "value": encapsulation_value}

        if parent_type == "node_group":
            parent_match = self.get_l3out_node_group(node_group, l3out_object.details, fail_module=True)
            parent_path = "/l3outTemplate/l3outs/{0}/nodeGroups/{1}".format(l3out_object.index, parent_match.index if parent_match else "-")
        elif parent_type == "floating_svi":
            parent_match = self.get_l3out_floating_svi_interface(l3out_object.details, pod_id, node_id, encap, fail_module=True)
            parent_path = "/l3outTemplate/l3outs/{0}/floatingSviInterfaces/{1}".format(l3out_object.index, parent_match.index if parent_match else "-")
        elif parent_type == "routed":
            parent_match = self.get_l3out_routed_interface(l3out_object.details, pod_id, node_id, path, port_channel_uuid, fail_module=True)
            parent_path = "/l3outTemplate/l3outs/{0}/interfaces/{1}".format(l3out_object.index, parent_match.index if parent_match else "-")
        elif parent_type == "routed_sub":
            parent_match = self.get_l3out_routed_sub_interface(l3out_object.details, pod_id, node_id, path, port_channel_uuid, encap, fail_module=True)
            parent_path = "/l3outTemplate/l3outs/{0}/subInterfaces/{1}".format(l3out_object.index, parent_match.index if parent_match else "-")
        elif parent_type == "svi":
            parent_match = self.get_l3out_svi_interface(
                l3out_object.details, pod_id, node_id, path, encap, port_channel_uuid or virtual_port_channel_uuid, fail_module=True
            )
            parent_path = "/l3outTemplate/l3outs/{0}/sviInterfaces/{1}".format(l3out_object.index, parent_match.index if parent_match else "-")
        elif parent_type == "floating_svi_path_attributes":
            floating_svi_object = self.get_l3out_floating_svi_interface(l3out_object.details, pod_id, node_id, encap, True)
            parent_match = self.get_l3out_floating_svi_interface_path_attributes(floating_svi_object.details, domain_type, domain, fail_module=True)
            parent_path = "/l3outTemplate/l3outs/{0}/floatingSviInterfaces/{1}/svi/floatingPathAttributes/{2}".format(
                l3out_object.index, floating_svi_object.index, parent_match.index if parent_match else "-"
            )

        return parent_match, parent_path

    def set_parent_details_for_nested_object_in_l3out(self, parent_type, parent_object, update_object):
        """
        Set the parent details for a nested object in the L3Out object.
        :param parent_type: The type of the parent object -> Str
        :param parent_object: The parent object details -> Dict
        :param update_object: The object to update with parent details -> Dict
        """
        parent_output_prepend = ""
        if parent_type == "node_group":
            parent_output_prepend = "nodeGroup"
        elif parent_type == "floating_svi":
            parent_output_prepend = "floatingSviInterface"
        elif parent_type == "routed":
            parent_output_prepend = "routedInterface"
        elif parent_type == "routed_sub":
            parent_output_prepend = "routedSubInterface"
        elif parent_type == "svi":
            parent_output_prepend = "sviInterface"
        elif parent_type == "floating_svi_path_attributes":
            parent_output_prepend = "floatingPathAttributes"

        pod_id = parent_object.get("podID")
        if pod_id:
            update_object["{0}PodID".format(parent_output_prepend)] = pod_id

        node_id = parent_object.get("nodeID")
        if node_id:
            update_object["{0}NodeID".format(parent_output_prepend)] = node_id

        path_type = parent_object.get("pathType")
        if path_type:
            update_object["{0}PathType".format(parent_output_prepend)] = path_type

        encap = parent_object.get("encap")
        if encap:
            update_object["{0}Encap".format(parent_output_prepend)] = encap

        path_ref = parent_object.get("pathRef")
        if path_ref:
            # Add pathType and pathRef to the update_object in order to resolve the port_channel or virtual_port_channel
            update_object["pathType"] = path_type
            update_object["pathRef"] = path_ref
            self.update_config_with_port_channel_references(update_object)
            update_object.pop("pathType", None)
            update_object.pop("pathRef", None)
            update_object["{0}PathRef".format(parent_output_prepend)] = path_ref

        path = parent_object.get("path")
        if path:
            update_object["{0}Path".format(parent_output_prepend)] = path

        name = parent_object.get("name")
        if name:
            update_object["{0}Name".format(parent_output_prepend)] = name

        domain = parent_object.get("domain")
        if domain:
            update_object["{0}Domain".format(parent_output_prepend)] = domain

    def get_route_map_policy(self, uuid=None, name=None, template_object=None, fail_module=False):
        """
        Get the Route Map Policy for Route Control by UUID or Name.
        :param uuid: UUID of the Route Map Policy for Route Control to search for -> Str
        :param name: Name of the Route Map Policy for Route Control to search for -> Str
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                When the UUID | Name is existing in the search list -> Dict
                When the UUID | Name is not existing in the search list -> None
                When both UUID and Name are None, and the search list is not empty -> List[Dict]
                When both UUID and Name are None, and the search list is empty -> List[]
        """
        template_object = template_object if template_object else self.template
        match = template_object.get("tenantPolicyTemplate", {}).get("template", {}).get("routeMapPolicies", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs(
                "Route Map Policy for Route Control", match, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module
            )
        return match  # Query all objects

    def get_set_rule_policy_object(self, uuid=None, name=None, search_object=None, fail_module=False):
        """
        Get the Set Rule Policy by uuid or name.
        :param uuid: UUID of the Set Rule Policy to search for -> Str
        :param name: Name of the Set Rule Policy to search for -> Str
        :param search_object: The object to search in -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        if not search_object:
            search_object = self.template

        existing_objects = search_object.get("tenantPolicyTemplate", {}).get("template", {}).get("setRulePolicies", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs(
                "Set Rule Policy",
                existing_objects,
                [KVPair("uuid", uuid) if uuid else KVPair("name", name)],
                fail_module,
            )

        return existing_objects  # Query all objects

    def get_netflow_record(self, uuid=None, name=None, search_object=None, fail_module=False):
        """
        Get the NetFlow Record by uuid or name.
        :param uuid: UUID of the NetFlow Record to search for -> Str
        :param name: Name of the NetFlow Record to search for -> Str
        :param search_object: The object to search in -> Dict
        :param fail_module: When match is not found fail the ansible module -> Bool
        :return: Dict | None | List[Dict] | List[]: The processed result which could be:
                 When the UUID | Name is existing in the search list -> Dict
                 When the UUID | Name is not existing in the search list -> None
                 When both UUID and Name are None, and the search list is not empty -> List[Dict]
                 When both UUID and Name are None, and the search list is empty -> List[]
        """
        search_object = search_object if search_object else self.template
        match = search_object.get("tenantPolicyTemplate", {}).get("template", {}).get("netFlowRecords", [])
        if uuid or name:  # Query a specific object
            return self.get_object_by_key_value_pairs("NetFlow Record", match, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module)
        return match  # Query all objects



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Ansible module to manage CheckPoint Firewall (c) 2019
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = """
---
module: cp_mgmt_radius_server
short_description: Manages radius-server objects on Checkpoint over Web Services API
description:
  - Manages radius-server objects on Checkpoint devices including creating, updating and removing objects.
  - All operations are performed over Web Services API.
  - Available from R81.20 management version.
version_added: "5.0.0"
author: "Shiran Golzar (@chkp-shirango)"
options:
  name:
    description:
      - Object name.
    type: str
    required: True
  server:
    description:
      - The UID or Name of the host that is the RADIUS Server.
    type: str
  shared_secret:
    description:
      - The secret between the RADIUS server and the Security Gateway.
    type: str
  service:
    description:
      - The UID or Name of the Service to which the RADIUS server listens.
    type: str
  server_version:
    description:
      - The version can be either RADIUS Version 1.0, which is RFC 2138 compliant, and RADIUS Version 2.0 which is RFC 2865 compliant.
    type: str
    choices: ['RADIUS Ver. 1.0', 'RADIUS Ver. 2.0']
  protocol:
    description:
      - The type of authentication protocol that will be used when authenticating the user to the RADIUS server.
    type: str
    choices: ['PAP', 'MS_CHAP2']
  priority:
    description:
      - The priority of the RADIUS Server in case it is a member of a RADIUS Group.
    type: int
  accounting:
    description:
      - Accounting settings.
    type: dict
    suboptions:
      enable_ip_pool_management:
        description:
          - IP pool management, enables Accounting service.
        type: bool
      accounting_service:
        description:
          - The UID or Name of the the accounting interface to notify the server when users login and logout which will then lock and release the
            IP addresses that the server allocated to those users.
        type: str
  tags:
    description:
      - Collection of tag identifiers.
    type: list
    elements: str
  color:
    description:
      - Color of the object. Should be one of existing colors.
    type: str
    choices: ['aquamarine', 'black', 'blue', 'crete blue', 'burlywood', 'cyan', 'dark green', 'khaki', 'orchid', 'dark orange', 'dark sea green',
             'pink', 'turquoise', 'dark blue', 'firebrick', 'brown', 'forest green', 'gold', 'dark gold', 'gray', 'dark gray', 'light green', 'lemon chiffon',
             'coral', 'sea green', 'sky blue', 'magenta', 'purple', 'slate blue', 'violet red', 'navy blue', 'olive', 'orange', 'red', 'sienna', 'yellow']
  comments:
    description:
      - Comments string.
    type: str
  details_level:
    description:
      - The level of detail for some of the fields in the response can vary from showing only the UID value of the object to a fully detailed
        representation of the object.
    type: str
    choices: ['uid', 'standard', 'full']
  groups:
    description:
      - Collection of group identifiers.
    type: list
    elements: str
  ignore_warnings:
    description:
      - Apply changes ignoring warnings.
    type: bool
  ignore_errors:
    description:
      - Apply changes ignoring errors. You won't be able to publish such a changes. If ignore-warnings flag was omitted - warnings will also be ignored.
    type: bool
extends_documentation_fragment: check_point.mgmt.checkpoint_objects
"""

EXAMPLES = """
- name: add-radius-server
  cp_mgmt_radius_server:
    name: radServer
    server: hostRad
    shared_secret: '123'
    state: present

- name: set-radius-server
  cp_mgmt_radius_server:
    name: t4
    server: hostRadius
    state: present

- name: delete-radius-server
  cp_mgmt_radius_server:
    ignore_warnings: 'true'
    name: radiusServer
    state: absent
"""

RETURN = """
cp_mgmt_radius_server:
  description: The checkpoint object created or updated.
  returned: always, except when deleting the object.
  type: dict
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.check_point.mgmt.plugins.module_utils.checkpoint import checkpoint_argument_spec_for_objects, \
    api_call


def main():
    argument_spec = dict(
        name=dict(type='str', required=True),
        server=dict(type='str'),
        shared_secret=dict(type='str', no_log=True),
        service=dict(type='str'),
        server_version=dict(type='str', choices=['RADIUS Ver. 1.0', 'RADIUS Ver. 2.0']),
        protocol=dict(type='str', choices=['PAP', 'MS_CHAP2']),
        priority=dict(type='int'),
        accounting=dict(type='dict', options=dict(
            enable_ip_pool_management=dict(type='bool'),
            accounting_service=dict(type='str')
        )),
        tags=dict(type='list', elements='str'),
        color=dict(type='str', choices=['aquamarine', 'black', 'blue', 'crete blue', 'burlywood', 'cyan', 'dark green',
                                        'khaki', 'orchid', 'dark orange', 'dark sea green', 'pink', 'turquoise',
                                        'dark blue', 'firebrick', 'brown',
                                        'forest green', 'gold', 'dark gold', 'gray', 'dark gray', 'light green',
                                        'lemon chiffon', 'coral', 'sea green',
                                        'sky blue', 'magenta', 'purple', 'slate blue', 'violet red', 'navy blue',
                                        'olive', 'orange', 'red', 'sienna',
                                        'yellow']),
        comments=dict(type='str'),
        details_level=dict(type='str', choices=['uid', 'standard', 'full']),
        groups=dict(type='list', elements='str'),
        ignore_warnings=dict(type='bool'),
        ignore_errors=dict(type='bool')
    )
    argument_spec.update(checkpoint_argument_spec_for_objects)

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)
    api_call_object = 'radius-server'

    result = api_call(module, api_call_object)
    module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2022, James Livulpi
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: homectl
author:
  - "James Livulpi (@jameslivulpi)"
short_description: Manage user accounts with systemd-homed
version_added: 4.4.0
description:
  - Manages a user's home directory managed by systemd-homed.
notes:
  - This module uses L(passlib, https://pypi.org/project/passlib/) for password hashing when available,
    falling back to the Python C(crypt) module or L(legacycrypt, https://pypi.org/project/legacycrypt/).
requirements:
  - passlib (Python library, recommended), or legacycrypt on Python 3.13 or newer
  - It requires no dependency on Python 3.12 and earlier, but then it relies on the deprecated standard library C(crypt).
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    description:
      - The user name to create, remove, or update.
    required: true
    aliases: ['user', 'username']
    type: str
  password:
    description:
      - Set the user's password to this.
      - Homed requires this value to be in cleartext on user creation and updating a user.
      - The module takes the password and generates a password hash in SHA-512 with 10000 rounds of salt generation using
        crypt.
      - See U(https://systemd.io/USER_RECORD/).
      - This is required for O(state=present). When an existing user is updated this is checked against the stored hash in
        homed.
    type: str
  state:
    description:
      - The operation to take on the user.
    choices: ['absent', 'present']
    default: present
    type: str
  storage:
    description:
      - Indicates the storage mechanism for the user's home directory.
      - If the storage type is not specified, C(homed.conf(5\)) defines which default storage to use.
      - Only used when a user is first created.
    choices: ['classic', 'luks', 'directory', 'subvolume', 'fscrypt', 'cifs']
    type: str
  disksize:
    description:
      - The intended home directory disk space.
      - Human readable value such as V(10G), V(10M), or V(10B).
    type: str
  resize:
    description:
      - When used with O(disksize) this attempts to resize the home directory immediately.
    default: false
    type: bool
  realname:
    description:
      - The user's real ('human') name.
      - This can also be used to add a comment to maintain compatibility with C(useradd).
    aliases: ['comment']
    type: str
  realm:
    description:
      - The 'realm' a user is defined in.
    type: str
  email:
    description:
      - The email address of the user.
    type: str
  location:
    description:
      - A free-form location string describing the location of the user.
    type: str
  iconname:
    description:
      - The name of an icon picked by the user, for example for the purpose of an avatar.
      - Should follow the semantics defined in the Icon Naming Specification.
      - See U(https://specifications.freedesktop.org/icon-naming-spec/icon-naming-spec-latest.html) for specifics.
    type: str
  homedir:
    description:
      - Path to use as home directory for the user.
      - This is the directory the user's home directory is mounted to while the user is logged in.
      - This is not where the user's data is actually stored, see O(imagepath) for that.
      - Only used when a user is first created.
    type: path
  imagepath:
    description:
      - Path to place the user's home directory.
      - See U(https://www.freedesktop.org/software/systemd/man/homectl.html#--image-path=PATH) for more information.
      - Only used when a user is first created.
    type: path
  uid:
    description:
      - Sets the UID of the user.
      - If using O(gid) homed requires the value to be the same.
      - Only used when a user is first created.
    type: int
  gid:
    description:
      - Sets the gid of the user.
      - If using O(uid) homed requires the value to be the same.
      - Only used when a user is first created.
    type: int
  mountopts:
    description:
      - String separated by comma each indicating mount options for a users home directory.
      - Valid options are V(nosuid), V(nodev) or V(noexec).
      - Homed by default uses V(nodev) and V(nosuid) while V(noexec) is off.
    type: str
  umask:
    description:
      - Sets the umask for the user's login sessions.
      - Value from V(0000) to V(0777).
    type: int
  memberof:
    description:
      - String separated by comma each indicating a UNIX group this user shall be a member of.
      - Groups the user should be a member of should be supplied as comma separated list.
    aliases: ['groups']
    type: str
  skeleton:
    description:
      - The absolute path to the skeleton directory to populate a new home directory from.
      - This is only used when a home directory is first created.
      - If not specified homed by default uses V(/etc/skel).
    aliases: ['skel']
    type: path
  shell:
    description:
      - Shell binary to use for terminal logins of given user.
      - If not specified homed by default uses V(/bin/bash).
    type: str
  environment:
    description:
      - String separated by comma each containing an environment variable and its value to set for the user's login session,
        in a format compatible with C(putenv(\)).
      - Any environment variable listed here is automatically set by pam_systemd for all login sessions of the user.
    aliases: ['setenv']
    type: str
  timezone:
    description:
      - Preferred timezone to use for the user.
      - Should be a tzdata compatible location string such as V(America/New_York).
    type: str
  locked:
    description:
      - Whether the user account should be locked or not.
    type: bool
  language:
    description:
      - The preferred language/locale for the user.
      - This should be in a format compatible with the E(LANG) environment variable.
    type: str
  passwordhint:
    description:
      - Password hint for the given user.
    type: str
  sshkeys:
    description:
      - String separated by comma each listing a SSH public key that is authorized to access the account.
      - The keys should follow the same format as the lines in a traditional C(~/.ssh/authorized_key) file.
    type: str
  notbefore:
    description:
      - A time since the UNIX epoch before which the record should be considered invalid for the purpose of logging in.
    type: int
  notafter:
    description:
      - A time since the UNIX epoch after which the record should be considered invalid for the purpose of logging in.
    type: int
"""

EXAMPLES = r"""
- name: Add the user 'james'
  community.general.homectl:
    name: johnd
    password: myreallysecurepassword1!
    state: present

- name: Add the user 'alice' with a zsh shell, uid of 1000, and gid of 2000
  community.general.homectl:
    name: alice
    password: myreallysecurepassword1!
    state: present
    shell: /bin/zsh
    uid: 1000
    gid: 1000

- name: Modify an existing user 'frank' to have 10G of diskspace and resize usage now
  community.general.homectl:
    name: frank
    password: myreallysecurepassword1!
    state: present
    disksize: 10G
    resize: true

- name: Remove an existing user 'janet'
  community.general.homectl:
    name: janet
    state: absent
"""

RETURN = r"""
data:
  description: Dictionary returned from C(homectl inspect -j).
  returned: success
  type: dict
  sample:
    {
      "data": {
        "binding": {
          "e9ed2a5b0033427286b228e97c1e8343": {
            "fileSystemType": "btrfs",
            "fileSystemUuid": "7bd59491-2812-4642-a492-220c3f0c6c0b",
            "gid": 60268,
            "imagePath": "/home/james.home",
            "luksCipher": "aes",
            "luksCipherMode": "xts-plain64",
            "luksUuid": "7f05825a-2c38-47b4-90e1-f21540a35a81",
            "luksVolumeKeySize": 32,
            "partitionUuid": "5a906126-d3c8-4234-b230-8f6e9b427b2f",
            "storage": "luks",
            "uid": 60268
          }
        },
        "diskSize": 3221225472,
        "disposition": "regular",
        "lastChangeUSec": 1641941238208691,
        "lastPasswordChangeUSec": 1641941238208691,
        "privileged": {
          "hashedPassword": [
            "$6$ov9AKni.trf76inT$tTtfSyHgbPTdUsG0CvSSQZXGqFGdHKQ9Pb6e0BTZhDmlgrL/vA5BxrXduBi8u/PCBiYUffGLIkGhApjKMK3bV."
          ]
        },
        "signature": [
          {
            "data": "o6zVFbymcmk4YTVaY6KPQK23YCp+VkXdGEeniZeV1pzIbFzoaZBvVLPkNKMoPAQbodY5BYfBtuy41prNL78qAg==",
            "key": "-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAbs7ELeiEYBxkUQhxZ+5NGyu6J7gTtZtZ5vmIw3jowcY=\n-----END PUBLIC KEY-----\n"
          }
        ],
        "status": {
          "e9ed2a5b0033427286b228e97c1e8343": {
            "diskCeiling": 21845405696,
            "diskFloor": 268435456,
            "diskSize": 3221225472,
            "service": "io.systemd.Home",
            "signedLocally": true,
            "state": "inactive"
          }
        },
        "userName": "james"
      }
    }
"""

import json

from ansible.module_utils.basic import AnsibleModule, jsonify
from ansible.module_utils.common.text.formatters import human_to_bytes

from ansible_collections.community.general.plugins.module_utils import _deps as deps

with deps.declare("crypt_context"):
    from ansible_collections.community.general.plugins.module_utils._crypt import CryptContext, has_crypt_context

    if not has_crypt_context:
        raise ImportError("Failed to import any of: passlib, crypt, legacycrypt")


class Homectl:
    def __init__(self, module, crypt_context):
        self.module = module
        self.crypt_context = crypt_context
        self.state = module.params["state"]
        self.name = module.params["name"]
        self.password = module.params["password"]
        self.storage = module.params["storage"]
        self.disksize = module.params["disksize"]
        self.resize = module.params["resize"]
        self.realname = module.params["realname"]
        self.realm = module.params["realm"]
        self.email = module.params["email"]
        self.location = module.params["location"]
        self.iconname = module.params["iconname"]
        self.homedir = module.params["homedir"]
        self.imagepath = module.params["imagepath"]
        self.uid = module.params["uid"]
        self.gid = module.params["gid"]
        self.umask = module.params["umask"]
        self.memberof = module.params["memberof"]
        self.skeleton = module.params["skeleton"]
        self.shell = module.params["shell"]
        self.environment = module.params["environment"]
        self.timezone = module.params["timezone"]
        self.locked = module.params["locked"]
        self.passwordhint = module.params["passwordhint"]
        self.sshkeys = module.params["sshkeys"]
        self.language = module.params["language"]
        self.notbefore = module.params["notbefore"]
        self.notafter = module.params["notafter"]
        self.mountopts = module.params["mountopts"]

        self.result = {}

    # Cannot run homectl commands if service is not active
    def homed_service_active(self):
        is_active = True
        cmd = ["systemctl", "show", "systemd-homed.service", "-p", "ActiveState"]
        rc, show_service_stdout, stderr = self.module.run_command(cmd)
        if rc == 0:
            state = show_service_stdout.rsplit("=")[1]
            if state.strip() != "active":
                is_active = False
        return is_active

    def user_exists(self):
        exists = False
        valid_pw = False
        # Get user properties if they exist in json
        rc, stdout, stderr = self.get_user_metadata()
        if rc == 0:
            exists = True
            # User exists now compare password given with current hashed password stored in the user metadata.
            if self.state != "absent":  # Don't need checking on remove user
                stored_pwhash = json.loads(stdout)["privileged"]["hashedPassword"][0]
                if self._check_password(stored_pwhash):
                    valid_pw = True
        return exists, valid_pw

    def create_user(self):
        record = self.create_json_record(create=True)
        cmd = [self.module.get_bin_path("homectl", True)]
        cmd.append("create")
        cmd.append("--identity=-")  # Read the user record from standard input.
        return self.module.run_command(cmd, data=record)

    def _hash_password(self, password):
        return self.crypt_context.hash(password, scheme="sha512_crypt", rounds=10000)

    def _check_password(self, pwhash):
        return self.crypt_context.verify(self.password, pwhash)

    def remove_user(self):
        cmd = [self.module.get_bin_path("homectl", True)]
        cmd.append("remove")
        cmd.append(self.name)
        return self.module.run_command(cmd)

    def prepare_modify_user_command(self):
        record = self.create_json_record()
        cmd = [self.module.get_bin_path("homectl", True)]
        cmd.append("update")
        cmd.append(self.name)
        cmd.append("--identity=-")  # Read the user record from standard input.
        # Resize disksize now resize = true
        # This is not valid in user record (json) and requires it to be passed on command.
        if self.disksize and self.resize:
            cmd.append("--and-resize")
            cmd.append("true")
            self.result["changed"] = True
        return cmd, record

    def get_user_metadata(self):
        cmd = [self.module.get_bin_path("homectl", True)]
        cmd.append("inspect")
        cmd.append(self.name)
        cmd.append("-j")
        cmd.append("--no-pager")
        rc, stdout, stderr = self.module.run_command(cmd)
        return rc, stdout, stderr

    # Build up dictionary to jsonify for homectl commands.
    def create_json_record(self, create=False):
        record = {}
        user_metadata = {}
        self.result["changed"] = False
        # Get the current user record if not creating a new user record.
        if not create:
            rc, user_metadata, stderr = self.get_user_metadata()
            user_metadata = json.loads(user_metadata)
            # Remove elements that are not meant to be updated from record.
            # These are always part of the record when a user exists.
            user_metadata.pop("signature", None)
            user_metadata.pop("binding", None)
            user_metadata.pop("status", None)
            # Let last change Usec be updated by homed when command runs.
            user_metadata.pop("lastChangeUSec", None)
            # Now only change fields that are called on leaving what's currently in the record intact.
            record = user_metadata

        record["userName"] = self.name
        record["secret"] = {"password": [self.password]}

        if create:
            password_hash = self._hash_password(self.password)
            record["privileged"] = {"hashedPassword": [password_hash]}
            self.result["changed"] = True

        if self.uid and self.gid and create:
            record["uid"] = self.uid
            record["gid"] = self.gid
            self.result["changed"] = True

        if self.memberof:
            member_list = list(self.memberof.split(","))
            if member_list != record.get("memberOf", [None]):
                record["memberOf"] = member_list
                self.result["changed"] = True

        if self.realname:
            if self.realname != record.get("realName"):
                record["realName"] = self.realname
                self.result["changed"] = True

        # Cannot update storage unless were creating a new user.
        # See 'Fields in the binding section' at https://systemd.io/USER_RECORD/
        if self.storage and create:
            record["storage"] = self.storage
            self.result["changed"] = True

        # Cannot update homedir unless were creating a new user.
        # See 'Fields in the binding section' at https://systemd.io/USER_RECORD/
        if self.homedir and create:
            record["homeDirectory"] = self.homedir
            self.result["changed"] = True

        # Cannot update imagepath unless were creating a new user.
        # See 'Fields in the binding section' at https://systemd.io/USER_RECORD/
        if self.imagepath and create:
            record["imagePath"] = self.imagepath
            self.result["changed"] = True

        if self.disksize:
            # convert human readable to bytes
            if self.disksize != record.get("diskSize"):
                record["diskSize"] = human_to_bytes(self.disksize)
                self.result["changed"] = True

        if self.realm:
            if self.realm != record.get("realm"):
                record["realm"] = self.realm
                self.result["changed"] = True

        if self.email:
            if self.email != record.get("emailAddress"):
                record["emailAddress"] = self.email
                self.result["changed"] = True

        if self.location:
            if self.location != record.get("location"):
                record["location"] = self.location
                self.result["changed"] = True

        if self.iconname:
            if self.iconname != record.get("iconName"):
                record["iconName"] = self.iconname
                self.result["changed"] = True

        if self.skeleton:
            if self.skeleton != record.get("skeletonDirectory"):
                record["skeletonDirectory"] = self.skeleton
                self.result["changed"] = True

        if self.shell:
            if self.shell != record.get("shell"):
                record["shell"] = self.shell
                self.result["changed"] = True

        if self.umask:
            if self.umask != record.get("umask"):
                record["umask"] = self.umask
                self.result["changed"] = True

        if self.environment:
            if self.environment != record.get("environment", [None]):
                record["environment"] = list(self.environment.split(","))
                self.result["changed"] = True

        if self.timezone:
            if self.timezone != record.get("timeZone"):
                record["timeZone"] = self.timezone
                self.result["changed"] = True

        if self.locked:
            if self.locked != record.get("locked"):
                record["locked"] = self.locked
                self.result["changed"] = True

        if self.passwordhint:
            if self.passwordhint != record.get("privileged", {}).get("passwordHint"):
                record["privileged"]["passwordHint"] = self.passwordhint
                self.result["changed"] = True

        if self.sshkeys:
            if self.sshkeys != record.get("privileged", {}).get("sshAuthorizedKeys"):
                record["privileged"]["sshAuthorizedKeys"] = list(self.sshkeys.split(","))
                self.result["changed"] = True

        if self.language:
            if self.locked != record.get("preferredLanguage"):
                record["preferredLanguage"] = self.language
                self.result["changed"] = True

        if self.notbefore:
            if self.locked != record.get("notBeforeUSec"):
                record["notBeforeUSec"] = self.notbefore
                self.result["changed"] = True

        if self.notafter:
            if self.locked != record.get("notAfterUSec"):
                record["notAfterUSec"] = self.notafter
                self.result["changed"] = True

        if self.mountopts:
            opts = list(self.mountopts.split(","))
            if "nosuid" in opts:
                if record.get("mountNoSuid") is not True:
                    record["mountNoSuid"] = True
                    self.result["changed"] = True
            else:
                if record.get("mountNoSuid") is not False:
                    record["mountNoSuid"] = False
                    self.result["changed"] = True

            if "nodev" in opts:
                if record.get("mountNoDevices") is not True:
                    record["mountNoDevices"] = True
                    self.result["changed"] = True
            else:
                if record.get("mountNoDevices") is not False:
                    record["mountNoDevices"] = False
                    self.result["changed"] = True

            if "noexec" in opts:
                if record.get("mountNoExecute") is not True:
                    record["mountNoExecute"] = True
                    self.result["changed"] = True
            else:
                if record.get("mountNoExecute") is not False:
                    record["mountNoExecute"] = False
                    self.result["changed"] = True

        return jsonify(record)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(type="str", default="present", choices=["absent", "present"]),
            name=dict(type="str", required=True, aliases=["user", "username"]),
            password=dict(type="str", no_log=True),
            storage=dict(type="str", choices=["classic", "luks", "directory", "subvolume", "fscrypt", "cifs"]),
            disksize=dict(type="str"),
            resize=dict(type="bool", default=False),
            realname=dict(type="str", aliases=["comment"]),
            realm=dict(type="str"),
            email=dict(type="str"),
            location=dict(type="str"),
            iconname=dict(type="str"),
            homedir=dict(type="path"),
            imagepath=dict(type="path"),
            uid=dict(type="int"),
            gid=dict(type="int"),
            umask=dict(type="int"),
            environment=dict(type="str", aliases=["setenv"]),
            timezone=dict(type="str"),
            memberof=dict(type="str", aliases=["groups"]),
            skeleton=dict(type="path", aliases=["skel"]),
            shell=dict(type="str"),
            locked=dict(type="bool"),
            passwordhint=dict(type="str", no_log=True),
            sshkeys=dict(type="str", no_log=True),
            language=dict(type="str"),
            notbefore=dict(type="int"),
            notafter=dict(type="int"),
            mountopts=dict(type="str"),
        ),
        supports_check_mode=True,
        required_if=[
            ("state", "present", ["password"]),
            ("resize", True, ["disksize"]),
        ],
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    deps.validate(module)

    crypt_context = CryptContext(schemes=["sha512_crypt", "sha256_crypt", "md5_crypt", "des_crypt"])
    homectl = Homectl(module, crypt_context)
    homectl.result["state"] = homectl.state

    # First we need to make sure homed service is active
    if not homectl.homed_service_active():
        module.fail_json(msg="systemd-homed.service is not active")

    # handle removing user
    if homectl.state == "absent":
        user_exists, valid_pwhash = homectl.user_exists()
        if user_exists:
            if module.check_mode:
                module.exit_json(changed=True)
            rc, stdout, stderr = homectl.remove_user()
            if rc != 0:
                module.fail_json(name=homectl.name, msg=stderr, rc=rc)
            homectl.result["changed"] = True
            homectl.result["rc"] = rc
            homectl.result["msg"] = f"User {homectl.name} removed!"
        else:
            homectl.result["changed"] = False
            homectl.result["msg"] = "User does not exist!"

    # Handle adding a user
    if homectl.state == "present":
        user_exists, valid_pwhash = homectl.user_exists()
        if not user_exists:
            if module.check_mode:
                module.exit_json(changed=True)
            rc, stdout, stderr = homectl.create_user()
            if rc != 0:
                module.fail_json(name=homectl.name, msg=stderr, rc=rc)
            rc, user_metadata, stderr = homectl.get_user_metadata()
            homectl.result["data"] = json.loads(user_metadata)
            homectl.result["rc"] = rc
            homectl.result["msg"] = f"User {homectl.name} created!"
        else:
            if valid_pwhash:
                # Run this to see if changed would be True or False which is useful for check_mode
                cmd, record = homectl.prepare_modify_user_command()
            else:
                # User gave wrong password fail with message
                homectl.result["changed"] = False
                homectl.result["msg"] = "User exists but password is incorrect!"
                module.fail_json(**homectl.result)

            if module.check_mode:
                module.exit_json(**homectl.result)

            # Now actually modify the user if changed was set to true at any point.
            if homectl.result["changed"]:
                rc, stdout, stderr = module.run_command(cmd, data=record)
                if rc != 0:
                    module.fail_json(name=homectl.name, msg=stderr, rc=rc, changed=False)
            rc, user_metadata, stderr = homectl.get_user_metadata()
            homectl.result["data"] = json.loads(user_metadata)
            homectl.result["rc"] = rc
            if homectl.result["changed"]:
                homectl.result["msg"] = f"User {homectl.name} modified"

    module.exit_json(**homectl.result)


if __name__ == "__main__":
    main()



# Copyright (c) 2022, Guillaume MARTINEZ <lunik@tiwabbit.fr>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:
    # Standard documentation fragment
    DOCUMENTATION = r"""
options:
  wait:
    description:
      - Wait for the resource to reach its desired state before returning.
    type: bool
    default: true
  wait_timeout:
    type: int
    description:
      - Time to wait for the resource to reach the expected state.
    required: false
    default: 300
  wait_sleep_time:
    type: int
    description:
      - Time to wait before every attempt to check the state of the resource.
    required: false
    default: 3
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: s3_sync
version_added: 1.0.0
short_description: Efficiently upload multiple files to S3
description:
- The S3 module is great, but it is very slow for a large volume of files- even a dozen will be noticeable. In addition to speed, it handles globbing,
  inclusions/exclusions, mime types, expiration mapping, recursion, cache control and smart directory mapping.
options:
  mode:
    description:
    - sync direction.
    default: 'push'
    choices: [ 'push' ]
    type: str
  file_change_strategy:
    description:
    - Difference determination method to allow changes-only syncing. Unlike rsync, files are not patched- they are fully skipped or fully uploaded.
    - date_size will upload if file sizes don't match or if local file modified date is newer than s3's version
    - checksum will compare etag values based on s3's implementation of chunked md5s.
    - force will always upload all files.
    required: false
    default: 'date_size'
    choices: [ 'force', 'checksum', 'date_size' ]
    type: str
  bucket:
    description:
    - Bucket name.
    required: true
    type: str
  key_prefix:
    description:
    - In addition to file path, prepend s3 path with this prefix. Module will add slash at end of prefix if necessary.
    required: false
    type: str
    default: ''
  file_root:
    description:
    - File/directory path for synchronization. This is a local path.
    - This root path is scrubbed from the key name, so subdirectories will remain as keys.
    required: true
    type: path
  permission:
    description:
    - Canned ACL to apply to synced files.
    - Changing this ACL only changes newly synced files, it does not trigger a full reupload.
    required: false
    choices:
    - 'private'
    - 'public-read'
    - 'public-read-write'
    - 'authenticated-read'
    - 'aws-exec-read'
    - 'bucket-owner-read'
    - 'bucket-owner-full-control'
    type: str
  mime_map:
    description:
    - >
      Dict entry from extension to MIME type. This will override any default/sniffed MIME type.
      For example C({".txt": "application/text", ".yml": "application/text"})
    required: false
    type: dict
  include:
    description:
    - Shell pattern-style file matching.
    - Used before exclude to determine eligible files (for instance, only C("*.gif"))
    - For multiple patterns, comma-separate them.
    required: false
    default: "*"
    type: str
  exclude:
    description:
    - Shell pattern-style file matching.
    - Used after include to remove files (for instance, skip C("*.txt"))
    - For multiple patterns, comma-separate them.
    required: false
    default: ".*"
    type: str
  cache_control:
    description:
    - Cache-Control header set on uploaded objects.
    - Directives are separated by commas.
    required: false
    type: str
    default: ''
  storage_class:
    description:
    - Storage class to be associated to each object added to the S3 bucket.
    required: false
    choices:
    - 'STANDARD'
    - 'REDUCED_REDUNDANCY'
    - 'STANDARD_IA'
    - 'ONEZONE_IA'
    - 'INTELLIGENT_TIERING'
    - 'GLACIER'
    - 'DEEP_ARCHIVE'
    - 'OUTPOSTS'
    default: 'STANDARD'
    type: str
    version_added: 1.5.0
  delete:
    description:
    - Remove remote files that exist in bucket but are not present in the file root.
    required: false
    default: false
    type: bool

author:
- Ted Timmons (@tedder)
extends_documentation_fragment:
- amazon.aws.common.modules
- amazon.aws.region.modules
- amazon.aws.boto3
"""

EXAMPLES = r"""
- name: basic upload
  community.aws.s3_sync:
    bucket: tedder
    file_root: roles/s3/files/

- name: basic upload using the glacier storage class
  community.aws.s3_sync:
    bucket: tedder
    file_root: roles/s3/files/
    storage_class: GLACIER

- name: basic individual file upload
  community.aws.s3_sync:
    bucket: tedder
    file_root: roles/s3/files/file_name

- name: all the options
  community.aws.s3_sync:
    bucket: tedder
    file_root: roles/s3/files
    mime_map:
      .yml: application/text
      .json: application/text
    key_prefix: config_files/web
    file_change_strategy: force
    permission: public-read
    cache_control: "public, max-age=31536000"
    storage_class: "GLACIER"
    include: "*"
    exclude: "*.txt,.*"
"""

RETURN = r"""
filelist_initial:
  description: file listing (dicts) from initial globbing
  returned: always
  type: list
  sample: [{
                "bytes": 151,
                "chopped_path": "policy.json",
                "fullpath": "roles/cf/files/policy.json",
                "modified_epoch": 1477416706
           }]
filelist_local_etag:
  description: file listing (dicts) including calculated local etag
  returned: always
  type: list
  sample: [{
                "bytes": 151,
                "chopped_path": "policy.json",
                "fullpath": "roles/cf/files/policy.json",
                "mime_type": "application/json",
                "modified_epoch": 1477416706,
                "s3_path": "s3sync/policy.json"
           }]
filelist_s3:
  description: file listing (dicts) including information about previously-uploaded versions
  returned: always
  type: list
  sample: [{
                "bytes": 151,
                "chopped_path": "policy.json",
                "fullpath": "roles/cf/files/policy.json",
                "mime_type": "application/json",
                "modified_epoch": 1477416706,
                "s3_path": "s3sync/policy.json"
           }]
filelist_typed:
  description: file listing (dicts) with calculated or overridden mime types
  returned: always
  type: list
  sample: [{
                "bytes": 151,
                "chopped_path": "policy.json",
                "fullpath": "roles/cf/files/policy.json",
                "mime_type": "application/json",
                "modified_epoch": 1477416706
           }]
filelist_actionable:
  description: file listing (dicts) of files that will be uploaded after the strategy decision
  returned: always
  type: list
  sample: [{
                "bytes": 151,
                "chopped_path": "policy.json",
                "fullpath": "roles/cf/files/policy.json",
                "mime_type": "application/json",
                "modified_epoch": 1477931256,
                "s3_path": "s3sync/policy.json",
                "whysize": "151 / 151",
                "whytime": "1477931256 / 1477929260"
           }]
uploads:
  description: file listing (dicts) of files that were actually uploaded
  returned: always
  type: list
  sample: [{
                "bytes": 151,
                "chopped_path": "policy.json",
                "fullpath": "roles/cf/files/policy.json",
                "s3_path": "s3sync/policy.json",
                "whysize": "151 / 151",
                "whytime": "1477931637 / 1477931489"
           }]

"""

import datetime
import fnmatch
import mimetypes
import os
import stat as osstat  # os.stat constants

try:
    from dateutil import tz

    HAS_DATEUTIL = True
except ImportError:
    HAS_DATEUTIL = False

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils._text import to_text

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code

from ansible_collections.community.aws.plugins.module_utils.etag import calculate_multipart_etag
from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


def gather_files(fileroot, include=None, exclude=None):
    ret = []

    if os.path.isfile(fileroot):
        fullpath = fileroot
        fstat = os.stat(fullpath)
        path_array = fileroot.split("/")
        chopped_path = path_array[-1]
        f_size = fstat[osstat.ST_SIZE]
        f_modified_epoch = fstat[osstat.ST_MTIME]
        ret.append(
            {
                "fullpath": fullpath,
                "chopped_path": chopped_path,
                "modified_epoch": f_modified_epoch,
                "bytes": f_size,
            }
        )

    else:
        for dirpath, dirnames, filenames in os.walk(fileroot):
            for fn in filenames:
                fullpath = os.path.join(dirpath, fn)
                # include/exclude
                if include:
                    found = False
                    for x in include.split(","):
                        if fnmatch.fnmatch(fn, x):
                            found = True
                    if not found:
                        # not on the include list, so we don't want it.
                        continue

                if exclude:
                    found = False
                    for x in exclude.split(","):
                        if fnmatch.fnmatch(fn, x):
                            found = True
                    if found:
                        # skip it, even if previously included.
                        continue

                chopped_path = os.path.relpath(fullpath, start=fileroot)
                fstat = os.stat(fullpath)
                f_size = fstat[osstat.ST_SIZE]
                f_modified_epoch = fstat[osstat.ST_MTIME]
                ret.append(
                    {
                        "fullpath": fullpath,
                        "chopped_path": chopped_path,
                        "modified_epoch": f_modified_epoch,
                        "bytes": f_size,
                    }
                )
            # dirpath = path *to* the directory
            # dirnames = subdirs *in* our directory
            # filenames
    return ret


def calculate_s3_path(filelist, key_prefix=""):
    ret = []
    for fileentry in filelist:
        # don't modify the input dict
        retentry = fileentry.copy()
        retentry["s3_path"] = os.path.join(key_prefix, fileentry["chopped_path"])
        ret.append(retentry)
    return ret


def calculate_local_etag(filelist, key_prefix=""):
    """Really, "calculate md5", but since AWS uses their own format, we'll just call
    it a "local etag". TODO optimization: only calculate if remote key exists."""
    ret = []
    for fileentry in filelist:
        # don't modify the input dict
        retentry = fileentry.copy()
        retentry["local_etag"] = calculate_multipart_etag(fileentry["fullpath"])
        ret.append(retentry)
    return ret


def determine_mimetypes(filelist, override_map):
    ret = []
    for fileentry in filelist:
        retentry = fileentry.copy()
        localfile = fileentry["fullpath"]

        # reminder: file extension is '.txt', not 'txt'.
        file_extension = os.path.splitext(localfile)[1]
        if override_map and override_map.get(file_extension):
            # override? use it.
            retentry["mime_type"] = override_map[file_extension]
        else:
            # else sniff it
            retentry["mime_type"], retentry["encoding"] = mimetypes.guess_type(localfile, strict=False)

        # might be None or '' from one of the above. Not a great type but better than nothing.
        if not retentry["mime_type"]:
            retentry["mime_type"] = "application/octet-stream"

        ret.append(retentry)

    return ret


def head_s3(s3, bucket, s3keys):
    retkeys = []
    for entry in s3keys:
        retentry = entry.copy()
        try:
            retentry["s3_head"] = s3.head_object(Bucket=bucket, Key=entry["s3_path"])
        # 404 (Missing) - File doesn't exist, we'll need to upload
        # 403 (Denied) - Sometimes we can write but not read, assume we'll need to upload
        except is_boto3_error_code(["404", "403"]):
            pass
        retkeys.append(retentry)
    return retkeys


def filter_list(s3, bucket, s3filelist, strategy):
    keeplist = list(s3filelist)

    for e in keeplist:
        e["_strategy"] = strategy

    # init/fetch info from S3 if we're going to use it for comparisons
    if not strategy == "force":
        keeplist = head_s3(s3, bucket, s3filelist)

    # now actually run the strategies
    if strategy == "checksum":
        for entry in keeplist:
            if entry.get("s3_head"):
                # since we have a remote s3 object, compare the values.
                if entry["s3_head"]["ETag"] == entry["local_etag"]:
                    # files match, so remove the entry
                    entry["skip_flag"] = True
                else:
                    # file etags don't match, keep the entry.
                    pass
            else:  # we don't have an etag, so we'll keep it.
                pass
    elif strategy == "date_size":
        for entry in keeplist:
            if entry.get("s3_head"):
                # fstat = entry['stat']
                local_modified_epoch = entry["modified_epoch"]
                local_size = entry["bytes"]

                # py2's datetime doesn't have a timestamp() field, so we have to revert to something more awkward.
                # remote_modified_epoch = entry['s3_head']['LastModified'].timestamp()
                remote_modified_datetime = entry["s3_head"]["LastModified"]
                delta = remote_modified_datetime - datetime.datetime(1970, 1, 1, tzinfo=tz.tzutc())
                remote_modified_epoch = delta.seconds + (delta.days * 86400)

                remote_size = entry["s3_head"]["ContentLength"]

                entry["whytime"] = f"{local_modified_epoch} / {remote_modified_epoch}"
                entry["whysize"] = f"{local_size} / {remote_size}"

                if local_modified_epoch <= remote_modified_epoch and local_size == remote_size:
                    entry["skip_flag"] = True
            else:
                entry["why"] = "no s3_head"
    # else: probably 'force'. Basically we don't skip with any with other strategies.
    else:
        pass

    # prune 'please skip' entries, if any.
    return [x for x in keeplist if not x.get("skip_flag")]


def upload_files(s3, bucket, filelist, params):
    ret = []
    for entry in filelist:
        args = {"ContentType": entry["mime_type"]}
        if params.get("permission"):
            args["ACL"] = params["permission"]
        if params.get("cache_control"):
            args["CacheControl"] = params["cache_control"]
        if params.get("storage_class"):
            args["StorageClass"] = params["storage_class"]
        # if this fails exception is caught in main()
        s3.upload_file(entry["fullpath"], bucket, entry["s3_path"], ExtraArgs=args, Callback=None, Config=None)
        ret.append(entry)
    return ret


def remove_files(s3, sourcelist, params):
    bucket = params.get("bucket")
    key_prefix = params.get("key_prefix")
    paginator = s3.get_paginator("list_objects_v2")
    current_keys = set(
        x["Key"] for x in paginator.paginate(Bucket=bucket, Prefix=key_prefix).build_full_result().get("Contents", [])
    )
    keep_keys = set(to_text(source_file["s3_path"]) for source_file in sourcelist)
    delete_keys = list(current_keys - keep_keys)

    # can delete 1000 objects at a time
    groups_of_keys = [delete_keys[i:i + 1000] for i in range(0, len(delete_keys), 1000)]  # fmt:skip
    for keys in groups_of_keys:
        s3.delete_objects(Bucket=bucket, Delete={"Objects": [{"Key": key} for key in keys]})

    return delete_keys


def main():
    argument_spec = dict(
        mode=dict(choices=["push"], default="push"),
        file_change_strategy=dict(choices=["force", "date_size", "checksum"], default="date_size"),
        bucket=dict(required=True),
        key_prefix=dict(required=False, default="", no_log=False),
        file_root=dict(required=True, type="path"),
        permission=dict(
            required=False,
            choices=[
                "private",
                "public-read",
                "public-read-write",
                "authenticated-read",
                "aws-exec-read",
                "bucket-owner-read",
                "bucket-owner-full-control",
            ],
        ),
        mime_map=dict(required=False, type="dict"),
        exclude=dict(required=False, default=".*"),
        include=dict(required=False, default="*"),
        cache_control=dict(required=False, default=""),
        delete=dict(required=False, type="bool", default=False),
        storage_class=dict(
            required=False,
            default="STANDARD",
            choices=[
                "STANDARD",
                "REDUCED_REDUNDANCY",
                "STANDARD_IA",
                "ONEZONE_IA",
                "INTELLIGENT_TIERING",
                "GLACIER",
                "DEEP_ARCHIVE",
                "OUTPOSTS",
            ],
        ),
        # future options: encoding, metadata, retries
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
    )

    if not HAS_DATEUTIL:
        module.fail_json(msg="dateutil required for this module")

    result = {}
    mode = module.params["mode"]

    try:
        s3 = module.client("s3")
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to connect to AWS")

    if mode == "push":
        try:
            result["filelist_initial"] = gather_files(
                module.params["file_root"], exclude=module.params["exclude"], include=module.params["include"]
            )
            result["filelist_typed"] = determine_mimetypes(result["filelist_initial"], module.params.get("mime_map"))
            result["filelist_s3"] = calculate_s3_path(result["filelist_typed"], module.params["key_prefix"])
            try:
                result["filelist_local_etag"] = calculate_local_etag(result["filelist_s3"])
            except ValueError as e:
                if module.params["file_change_strategy"] == "checksum":
                    module.fail_json_aws(
                        e,
                        "Unable to calculate checksum.  If running in FIPS mode, you may need to use another file_change_strategy",
                    )
                result["filelist_local_etag"] = result["filelist_s3"].copy()
            result["filelist_actionable"] = filter_list(
                s3, module.params["bucket"], result["filelist_local_etag"], module.params["file_change_strategy"]
            )
            result["uploads"] = upload_files(s3, module.params["bucket"], result["filelist_actionable"], module.params)

            if module.params["delete"]:
                result["removed"] = remove_files(s3, result["filelist_local_etag"], module.params)

            # mark changed if we actually upload something.
            if result.get("uploads") or result.get("removed"):
                result["changed"] = True
            # result.update(filelist=actionable_filelist)
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            module.fail_json_aws(e, msg="Failed to push file")

    module.exit_json(**result)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The ios_ospfv2 class
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to it's desired end-state is
created
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.facts.facts import Facts
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.rm_templates.ospfv2 import (
    Ospfv2Template,
)


class Ospfv2(ResourceModule):
    """
    The ios_ospfv2 class
    """

    def __init__(self, module):
        super(Ospfv2, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="ospfv2",
            tmplt=Ospfv2Template(),
        )

        self.parsers = [
            "adjacency",
            "address_family",
            "auto_cost",
            "bfd",
            "capability.lls",
            "capability.opaque",
            "capability.transit",
            "capability.vrf_lite",
            "compatible",
            "default_information",
            "default_metric",
            "discard_route",
            "distance.admin_distance",
            "distance.ospf",
            "distribute_list.prefix",
            "distribute_list.route_map",
            "domain_id",
            "domain_tag",
            "event_log",
            "help",
            "ignore",
            "interface_id",
            "ispf",
            "limit",
            "local_rib_criteria",
            "log_adjacency_changes",
            "max_lsa",
            "max_metric",
            "maximum_paths",
            "mpls.ldp.autoconfig",
            "mpls.ldp.sync",
            "mpls.traffic_eng",
            "neighbor",
            "nsf.cisco",
            "nsf.ietf.disable",
            "nsf.ietf.strict_lsa_checking",
            "prefix_suppression",
            "priority",
            "queue_depth.hello.max_packets",
            "queue_depth.hello.unlimited",
            "queue_depth.update.max_packets",
            "queue_depth.update.unlimited",
            "router_id",
            "shutdown",
            "summary_address.not_advertise",
            "summary_address.nssa_only",
            "timers.throttle.lsa",
            "timers.pacing.flood",
            "timers.pacing.lsa_group",
            "timers.pacing.retransmission",
            "timers.throttle.spf",
            "traffic_share",
            "ttl_security",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Select the appropriate function based on the state provided

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        haved, wantd = dict(), dict()

        if self.want:
            for entry in self.want.get("processes", []):
                entry = self._handle_deprecated(entry)
                wantd.update({(entry["process_id"], entry.get("vrf")): entry})

        if self.have:
            for entry in self.have.get("processes", []):
                # entry = self._handle_deprecated(entry)
                haved.update({(entry["process_id"], entry.get("vrf")): entry})

        # turn all lists of dicts into dicts prior to merge
        for each in wantd, haved:
            if each:
                self._list_to_dict(each)

        # if state is merged, merge want onto have
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, limit the have to anything in want
        # set want to nothing
        if self.state == "deleted":
            temp = {}
            for k, v in haved.items():
                if k in wantd or not wantd:
                    temp.update({k: v})
            haved = temp
            wantd = {}

        # delete processes first so we do run into "more than one" errors
        if self.state in ["overridden", "deleted"]:
            for k, have in haved.items():
                if k not in wantd:
                    self.addcmd(have, "pid", True)

        for k, want in wantd.items():
            self._compare(want=want, have=haved.pop(k, {}))

    def _compare(self, want, have):
        if want != have:
            self.addcmd(want or have, "pid", False)
            self.compare(self.parsers, want, have)
            self._areas_compare(want, have)
            self._complex_compare(want, have)
            if want.get("passive_interfaces"):
                self._passive_interfaces_compare(want, have)

    def _complex_compare(self, want, have):
        complex_parsers = ["distribute_list.acls", "network"]
        for _parser in complex_parsers:
            if _parser == "distribute_list.acls":
                wdist = want.get("distribute_list", {}).get("acls", {})
                hdist = have.get("distribute_list", {}).get("acls", {})
            else:
                wdist = want.get(_parser, {})
                hdist = have.get(_parser, {})
            for key, wanting in wdist.items():
                haveing = hdist.pop(key, {})
                if wanting != haveing:
                    if haveing and self.state in ["overridden", "replaced"]:
                        self.addcmd(haveing, _parser, negate=True)
                    self.addcmd(wanting, _parser, False)
            for key, haveing in hdist.items():
                self.addcmd(haveing, _parser, negate=True)

    def _areas_compare(self, want, have):
        wareas = want.get("areas", {})
        hareas = have.get("areas", {})
        for name, entry in wareas.items():
            self._area_compare(want=entry, have=hareas.pop(name, {}))
        for name, entry in hareas.items():
            self._area_compare(want={}, have=entry)

    def _area_compare(self, want, have):
        parsers = [
            "authentication",
            "capability",
            "default_cost",
            "nssa",
            "nssa.translate",
            "sham_link",
            "stub",
        ]
        self.compare(parsers=parsers, want=want, have=have)
        self._area_complex_compare(want, have, want.get("area_id"))

    def _area_complex_compare(self, want, have, area_id):
        area_complex_parsers = ["filter_list", "ranges"]
        for _parser in area_complex_parsers:
            wantr = want.get(_parser, {})
            haver = have.get(_parser, {})
            for key, wanting in wantr.items():
                haveing = have.pop(key, {})
                haveing["area_id"] = area_id
                wanting["area_id"] = area_id
                if wanting != haveing:
                    if haveing and self.state in ["overridden", "replaced"]:
                        self.addcmd(haveing, _parser, negate=True)
                    self.addcmd(wanting, _parser, False)
            for key, haveing in haver.items():
                haveing["area_id"] = area_id
                self.addcmd(haveing, _parser, negate=True)

    def _passive_interfaces_compare(self, want, have):
        parsers = ["passive_interfaces.default", "passive_interfaces.interface"]
        h_pi = None
        for k, v in want["passive_interfaces"].items():
            h_pi = have.get("passive_interfaces", {})
            if h_pi.get(k) and h_pi.get(k) != v:
                for each in v["name"]:
                    h_interface_name = h_pi[k].get("name", [])
                    if each not in h_interface_name:
                        temp = {"interface": {each: each}, "set_interface": v["set_interface"]}
                        self.compare(
                            parsers=parsers,
                            want={"passive_interfaces": temp},
                            have=dict(),
                        )
                    else:
                        h_interface_name.pop(each)
            elif not h_pi:
                if k == "interface":
                    for each in v["name"]:
                        temp = {"interface": {each: each}, "set_interface": v["set_interface"]}
                        self.compare(
                            parsers=parsers,
                            want={"passive_interfaces": temp},
                            have=dict(),
                        )
                elif k == "default":
                    self.compare(
                        parsers=parsers,
                        want={"passive_interfaces": {"default": True}},
                        have=dict(),
                    )
            else:
                h_pi.pop(k)
        if (self.state == "replaced" or self.state == "overridden") and h_pi:
            if h_pi.get("default") or h_pi.get("interface"):
                for k, v in h_pi.items():
                    if k == "interface":
                        for each in v["name"]:
                            temp = {
                                "interface": {each: each},
                                "set_interface": not (v["set_interface"]),
                            }
                            self.compare(
                                parsers=parsers,
                                want={"passive_interfaces": temp},
                                have=dict(),
                            )
                    elif k == "default":
                        self.compare(
                            parsers=parsers,
                            want=dict(),
                            have={"passive_interfaces": {"default": True}},
                        )

    def _list_to_dict(self, param):
        for _pid, proc in param.items():
            # convert list to dict for areas
            for area in proc.get("areas", []):
                area["ranges"] = {entry["address"]: entry for entry in area.get("ranges", [])}
                area["filter_list"] = {
                    entry["direction"]: entry for entry in area.get("filter_list", [])
                }

            proc["areas"] = {entry["area_id"]: entry for entry in proc.get("areas", [])}

            # list to dict for distribute_list
            distribute_list = proc.get("distribute_list", {})
            if "acls" in distribute_list:
                distribute_list["acls"] = {
                    entry["name"]: entry for entry in distribute_list["acls"]
                }

            # list to dict for passive_interfaces
            passive_interfaces = proc.get("passive_interfaces", {}).get("interface", {})
            if passive_interfaces.get("name"):
                passive_interfaces["name"] = {entry: entry for entry in passive_interfaces["name"]}

            # list to dict for network
            if proc.get("network"):
                proc["network"] = {entry["address"]: entry for entry in proc["network"]}

    def _handle_deprecated(self, config):
        if config.get("passive_interface"):
            passive_interfaces = config.get("passive_interfaces", {})
            interface = passive_interfaces.get("interface", {})
            name_list = interface.get("name", [])
            if not name_list:
                name_list.append(config["passive_interface"])
            else:
                name_list.extend(config["passive_interface"])
            del config["passive_interface"]
        return config



# Copyright (c) 2020, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import base64
import typing as t

from ansible.module_utils.common.text.converters import to_bytes

from ansible_collections.community.crypto.plugins.module_utils._crypto.basic import (
    OpenSSLObjectError,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.module_backends.privatekey import (
    get_privatekey_argument_spec,
    select_backend,
)
from ansible_collections.community.crypto.plugins.plugin_utils._action_module import (
    ActionModuleBase,
)

if t.TYPE_CHECKING:
    from ansible_collections.community.crypto.plugins.module_utils._argspec import (  # pragma: no cover
        ArgumentSpec,
    )
    from ansible_collections.community.crypto.plugins.module_utils._crypto.module_backends.privatekey import (  # pragma: no cover
        PrivateKeyBackend,
    )
    from ansible_collections.community.crypto.plugins.plugin_utils._action_module import (  # pragma: no cover
        AnsibleActionModule,
    )


class PrivateKeyModule:
    def __init__(
        self, module: AnsibleActionModule, module_backend: PrivateKeyBackend
    ) -> None:
        self.module = module
        self.module_backend = module_backend
        self.check_mode = module.check_mode
        self.changed = False
        self.return_current_key: bool = module.params["return_current_key"]

        content: str | None = module.params["content"]
        content_base64: bool = module.params["content_base64"]
        if content is not None:
            if content_base64:
                try:
                    data = base64.b64decode(content)
                except Exception as e:
                    module.fail_json(msg=f"Cannot decode Base64 encoded data: {e}")
            else:
                data = to_bytes(content)
            module_backend.set_existing(privatekey_bytes=data)

    def generate(self, module: AnsibleActionModule) -> None:
        """Generate a keypair."""

        if self.module_backend.needs_regeneration():
            # Regenerate
            self.module_backend.generate_private_key()
            # Call get_private_key_data() to make sure that exceptions are raised now:
            self.module_backend.get_private_key_data()
            self.changed = True
        elif self.module_backend.needs_conversion():
            # Convert
            self.module_backend.convert_private_key()
            # Call get_private_key_data() to make sure that exceptions are raised now:
            self.module_backend.get_private_key_data()
            self.changed = True

    def dump(self) -> dict[str, t.Any]:
        """Serialize the object into a dictionary."""
        result = self.module_backend.dump(
            include_key=self.changed or self.return_current_key
        )
        result["changed"] = self.changed
        return result


class ActionModule(ActionModuleBase):
    def setup_module(self) -> tuple[ArgumentSpec, dict[str, t.Any]]:
        argument_spec = get_privatekey_argument_spec()
        argument_spec.argument_spec.update(
            {
                "content": {"type": "str", "no_log": True},
                "content_base64": {"type": "bool", "default": False},
                "return_current_key": {"type": "bool", "default": False},
            }
        )
        return argument_spec, {
            "supports_check_mode": True,
        }

    def run_module(self, module: AnsibleActionModule) -> None:
        module_backend = select_backend(module=module)

        try:
            private_key = PrivateKeyModule(module, module_backend)
            private_key.generate(module)
            result = private_key.dump()
            if private_key.return_current_key:
                # In case the module's input (`content`) is returned as `privatekey`:
                # Since `content` is no_log=True, `privatekey`'s value will get replaced by
                # VALUE_SPECIFIED_IN_NO_LOG_PARAMETER. To avoid this, we remove the value of
                # `content` from module.no_log_values. Since we explicitly set
                # `module.no_log = True`, this should be safe.
                module.no_log = True
                try:
                    module.no_log_values.remove(module.params["content"])
                except KeyError:
                    pass
                module.params["content"] = "ANSIBLE_NO_LOG_VALUE"
            module.exit_json(**result)
        except OpenSSLObjectError as exc:
            module.fail_json(msg=str(exc))



# Copyright (c) 2016 Michael Gruener <michael.gruener@chaosmoon.net>
# Copyright (c) 2021 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import abc
import datetime
import re
import typing as t

from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    BackendException,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.basic import (
    OpenSSLObjectError,
)
from ansible_collections.community.crypto.plugins.module_utils._time import (
    UTC,
    ensure_utc_timezone,
    from_epoch_seconds,
    get_epoch_seconds,
    get_now_datetime,
    get_relative_time_option,
    remove_timezone,
)

if t.TYPE_CHECKING:
    import os  # pragma: no cover

    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover

    from ansible_collections.community.crypto.plugins.module_utils._acme.certificates import (  # pragma: no cover
        ChainMatcher,
        Criterium,
    )


class CertificateInformation(t.NamedTuple):
    not_valid_after: datetime.datetime
    not_valid_before: datetime.datetime
    serial_number: int
    subject_key_identifier: bytes | None
    authority_key_identifier: bytes | None


_FRACTIONAL_MATCHER = re.compile(
    r"^(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2})(|\.\d+)(Z|[+-]\d{2}:?\d{2}.*)$"
)


def _reduce_fractional_digits(timestamp_str: str) -> str:
    """
    Given a RFC 3339 timestamp that includes too many digits for the fractional seconds part, reduces these to at most 6.
    """
    # RFC 3339 (https://www.rfc-editor.org/info/rfc3339)
    m = _FRACTIONAL_MATCHER.match(timestamp_str)
    if not m:
        raise BackendException(f"Cannot parse ISO 8601 timestamp {timestamp_str!r}")
    timestamp, fractional, timezone = m.groups()
    if len(fractional) > 7:
        # Python does not support anything smaller than microseconds
        # (Golang supports nanoseconds, Boulder often emits more fractional digits, which Python chokes on)
        fractional = fractional[:7]
    return f"{timestamp}{fractional}{timezone}"


def _parse_acme_timestamp(
    timestamp_str: str, *, with_timezone: bool
) -> datetime.datetime:
    """
    Parses a RFC 3339 timestamp.
    """
    # RFC 3339 (https://www.rfc-editor.org/info/rfc3339)
    timestamp_str = _reduce_fractional_digits(timestamp_str)
    for time_format in (
        "%Y-%m-%dT%H:%M:%SZ",
        "%Y-%m-%dT%H:%M:%S.%fZ",
        "%Y-%m-%dT%H:%M:%S%z",
        "%Y-%m-%dT%H:%M:%S.%f%z",
    ):
        try:
            result = datetime.datetime.strptime(timestamp_str, time_format)
        except ValueError:
            pass
        else:
            return (
                ensure_utc_timezone(result)
                if with_timezone
                else remove_timezone(result)
            )
    raise BackendException(f"Cannot parse ISO 8601 timestamp {timestamp_str!r}")


class CryptoBackend(metaclass=abc.ABCMeta):
    def __init__(self, *, module: AnsibleModule, with_timezone: bool = False) -> None:
        self.module = module
        self._with_timezone = with_timezone

    def get_now(self) -> datetime.datetime:
        return get_now_datetime(with_timezone=self._with_timezone)

    def parse_acme_timestamp(self, timestamp_str: str) -> datetime.datetime:
        # RFC 3339 (https://www.rfc-editor.org/info/rfc3339)
        return _parse_acme_timestamp(timestamp_str, with_timezone=self._with_timezone)

    def parse_module_parameter(self, *, value: str, name: str) -> datetime.datetime:
        try:
            result = get_relative_time_option(
                value, input_name=name, with_timezone=self._with_timezone
            )
            if result is None:
                raise BackendException(f"Invalid value for {name}: {value!r}")
            return result
        except OpenSSLObjectError as exc:
            raise BackendException(str(exc)) from exc

    def interpolate_timestamp(
        self,
        timestamp_start: datetime.datetime,
        timestamp_end: datetime.datetime,
        *,
        percentage: float,
    ) -> datetime.datetime:
        start = get_epoch_seconds(timestamp_start)
        end = get_epoch_seconds(timestamp_end)
        return from_epoch_seconds(
            start + percentage * (end - start), with_timezone=self._with_timezone
        )

    def get_utc_datetime(
        self,
        year: int,
        month: int,
        day: int,
        hour: int = 0,
        minute: int = 0,
        second: int = 0,
        microsecond: int = 0,
        tzinfo: datetime.timezone | None = None,
    ) -> datetime.datetime:
        has_tzinfo = tzinfo is not None
        if self._with_timezone and not has_tzinfo:
            tzinfo = UTC
        result = datetime.datetime(
            year, month, day, hour, minute, second, microsecond, tzinfo
        )
        if self._with_timezone and has_tzinfo:
            result = ensure_utc_timezone(result)
        return result

    @abc.abstractmethod
    def parse_key(
        self,
        *,
        key_file: str | os.PathLike | None = None,
        key_content: str | None = None,
        passphrase: str | None = None,
    ) -> dict[str, t.Any]:
        """
        Parses an RSA or Elliptic Curve key file in PEM format and returns key_data.
        Raises KeyParsingError in case of errors.
        """

    @abc.abstractmethod
    def sign(
        self, *, payload64: str, protected64: str, key_data: dict[str, t.Any]
    ) -> dict[str, t.Any]:
        pass

    @abc.abstractmethod
    def create_mac_key(self, *, alg: str, key: str) -> dict[str, t.Any]:
        """Create a MAC key."""

    @abc.abstractmethod
    def get_ordered_csr_identifiers(
        self,
        *,
        csr_filename: str | os.PathLike | None = None,
        csr_content: str | bytes | None = None,
    ) -> list[tuple[str, str]]:
        """
        Return a list of requested identifiers (CN and SANs) for the CSR.
        Each identifier is a pair (type, identifier), where type is either
        'dns' or 'ip'.

        The list is deduplicated, and if a CNAME is present, it will be returned
        as the first element in the result.
        """

    @abc.abstractmethod
    def get_csr_identifiers(
        self,
        *,
        csr_filename: str | os.PathLike | None = None,
        csr_content: str | bytes | None = None,
    ) -> set[tuple[str, str]]:
        """
        Return a set of requested identifiers (CN and SANs) for the CSR.
        Each identifier is a pair (type, identifier), where type is either
        'dns' or 'ip'.
        """

    @abc.abstractmethod
    def get_cert_days(
        self,
        *,
        cert_filename: str | os.PathLike | None = None,
        cert_content: str | bytes | None = None,
        now: datetime.datetime | None = None,
    ) -> int:
        """
        Return the days the certificate in cert_filename remains valid and -1
        if the file was not found. If cert_filename contains more than one
        certificate, only the first one will be considered.

        If now is not specified, datetime.datetime.now() is used.
        """

    @abc.abstractmethod
    def create_chain_matcher(self, *, criterium: Criterium) -> ChainMatcher:
        """
        Given a Criterium object, creates a ChainMatcher object.
        """

    @abc.abstractmethod
    def get_cert_information(
        self,
        *,
        cert_filename: str | os.PathLike | None = None,
        cert_content: str | bytes | None = None,
    ) -> CertificateInformation:
        """
        Return some information on a X.509 certificate as a CertificateInformation object.
        """


__all__ = ("CertificateInformation", "CryptoBackend")
